from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Functions for creating RPE GateSets and GateString lists """
import numpy as _np
from . import gatesetconstruction as _setc
from . import datasetconstruction as _dsc

from .. import objects as _objs
from .. import tools as _tools


def make_parameterized_rpe_gate_set(alphaTrue, epsilonTrue, Yrot, SPAMdepol,
                                   gateDepol=None, withId=True):
    """
    Make a gateset for simulating RPE, paramaterized by rotation angles.  Note
    that the output gateset also has thetaTrue, alphaTrue, and epsilonTrue
    added attributes.

    Parameters
    ----------
    alphaTrue : float
       Angle of Z rotation (canonical RPE requires alphaTrue to be close to
       pi/2).

    epsilonTrue : float
       Angle of X rotation (canonical RPE requires epsilonTrue to be close to
       pi/4).

    Yrot : float
       Angle of rotation about Y axis that, by similarity transformation,
       rotates X rotation.

    SPAMdepol : float
       Amount to depolarize SPAM by.

    gateDepol : float, optional
       Amount to depolarize gates by (defaults to None).

    withId : bool, optional
       Do we include (perfect) identity or no identity? (Defaults to False;
       should be False for RPE, True for GST)

    Returns
    -------
    GateSet
        The desired gateset for RPE; gateset also has attributes thetaTrue,
        alphaTrue, and epsilonTrue, automatically extracted.
    """

    if withId:
        outputGateset = _setc.build_gateset(
            [2], [('Q0',)],['Gi','Gx','Gz'],
            [ "I(Q0)", "X(%s,Q0)" % epsilonTrue, "Z(%s,Q0)" % alphaTrue],
            prepLabels=["rho0"], prepExpressions=["0"],
            effectLabels=["E0"], effectExpressions=["1"],
            spamdefs={'plus': ('rho0','E0'), 'minus': ('rho0','remainder') } )
    else:
        outputGateset = _setc.build_gateset(
            [2], [('Q0',)],['Gx','Gz'],
            [ "X(%s,Q0)" % epsilonTrue, "Z(%s,Q0)" % alphaTrue],
            prepLabels=["rho0"], prepExpressions=["0"],
            effectLabels=["E0"], effectExpressions=["1"],
            spamdefs={'plus': ('rho0','E0'), 'minus': ('rho0','remainder') } )

    if Yrot != 0:
        gatesetAux1 = _setc.build_gateset(
            [2], [('Q0',)],['Gi','Gy','Gz'],
            [ "I(Q0)", "Y(%s,Q0)" % Yrot, "Z(pi/2,Q0)"],
            prepLabels=["rho0"], prepExpressions=["0"],
            effectLabels=["E0"], effectExpressions=["1"],
            spamdefs={'plus': ('rho0','E0'), 'minus': ('rho0','remainder') } )

        outputGateset.gates['Gx'] = _objs.FullyParameterizedGate(
            _np.dot( _np.dot(_np.linalg.inv(gatesetAux1.gates['Gy']),
                             outputGateset.gates['Gx']),gatesetAux1.gates['Gy']))

    outputGateset = outputGateset.depolarize(gate_noise=gateDepol,
                                             spam_noise=SPAMdepol)

    thetaTrue = _tools.rpe.extract_theta(outputGateset)
    outputGateset.thetaTrue = thetaTrue

    outputGateset.alphaTrue = _tools.rpe.extract_alpha(outputGateset)
    outputGateset.alphaTrue = alphaTrue

    outputGateset.epsilonTrue = _tools.rpe.extract_epsilon(outputGateset)
    outputGateset.epsilonTrue = epsilonTrue

    return outputGateset

def make_rpe_alpha_str_lists_gx_gz(kList):
    """
    Make alpha cosine and sine gatestring lists for (approx) X pi/4 and Z pi/2
    gates. These gate strings are used to estimate alpha (Z rotation angle).

    Parameters
    ----------
    kList : list of ints
       The list of "germ powers" to be used.  Typically successive powers of
       two; e.g. [1,2,4,8,16].

    Returns
    -------
    cosStrList : list of GateStrings
        The list of "cosine strings" to be used for alpha estimation.
    sinStrList : list of GateStrings
        The list of "sine strings" to be used for alpha estimation.
    """
    cosStrList = []
    sinStrList = []
    for k in kList:
        cosStrList += [ _objs.GateString(('Gi','Gx','Gx','Gz')+
                                         ('Gz',)*k +
                                         ('Gz','Gz','Gz','Gx','Gx'),
                                         'GiGxGxGzGz^'+str(k)+'GzGzGzGxGx')]

        sinStrList += [ _objs.GateString(('Gx','Gx','Gz','Gz')+
                                         ('Gz',)*k +
                                         ('Gz','Gz','Gz','Gx','Gx'),
                                         'GxGxGzGzGz^'+str(k)+'GzGzGzGxGx')]

        #From RPEToolsNewNew.py
        ##cosStrList += [_objs.GateString(('Gi','Gx','Gx')+
        ##                                ('Gz',)*k +
        ##                                ('Gx','Gx'),
        ##                                'GiGxGxGz^'+str(k)+'GxGx')]
        #
        #
        #cosStrList += [_objs.GateString(('Gx','Gx')+
        #                                ('Gz',)*k +
        #                                ('Gx','Gx'),
        #                                'GxGxGz^'+str(k)+'GxGx')]
        #
        #
        #sinStrList += [_objs.GateString(('Gx','Gx')+
        #                                ('Gz',)*k +
        #                                ('Gz','Gx','Gx'),
        #                                'GxGxGz^'+str(k)+'GzGxGx')]

    return cosStrList, sinStrList

def make_rpe_epsilon_str_lists_gx_gz(kList):
    """
    Make epsilon cosine and sine gatestring lists for (approx) X pi/4 and
    Z pi/2 gates. These gate strings are used to estimate epsilon (X rotation
    angle).

    Parameters
    ----------
    kList : list of ints
       The list of "germ powers" to be used.  Typically successive powers of
       two; e.g. [1,2,4,8,16].

    Returns
    -------
    epsilonCosStrList : list of GateStrings
        The list of "cosine strings" to be used for epsilon estimation.
    epsilonSinStrList : list of GateStrings
        The list of "sine strings" to be used for epsilon estimation.
    """
    epsilonCosStrList = []
    epsilonSinStrList = []

    for k in kList:
        epsilonCosStrList += [_objs.GateString(('Gx',)*k+
                                               ('Gx',)*4,
                                               'Gx^'+str(k)+'GxGxGxGx')]

        epsilonSinStrList += [_objs.GateString(('Gx','Gx','Gz','Gz')+
                                               ('Gx',)*k+
                                               ('Gx',)*4,
                                               'GxGxGzGzGx^'+str(k)+'GxGxGxGx')]

        #From RPEToolsNewNew.py
        #epsilonCosStrList += [_objs.GateString(('Gx',)*k,
        #                                       'Gx^'+str(k))]
        #
        #epsilonSinStrList += [_objs.GateString(('Gx','Gx')+('Gx',)*k,
        #                                       'GxGxGx^'+str(k))]

    return epsilonCosStrList, epsilonSinStrList

def make_rpe_theta_str_lists_gx_gz(kList):
    """
    Make theta cosine and sine gatestring lists for (approx) X pi/4 and Z pi/2
    gates. These gate strings are used to estimate theta (X-Z axes angle).

    Parameters
    ----------
    kList : list of ints
       The list of "germ powers" to be used.  Typically successive powers of
       two; e.g. [1,2,4,8,16].

    Returns
    -------
    thetaCosStrList : list of GateStrings
        The list of "cosine strings" to be used for theta estimation.
    thetaSinStrList : list of GateStrings
        The list of "sine strings" to be used for theta estimation.
    """
    thetaCosStrList = []
    thetaSinStrList = []

    for k in kList:
        thetaCosStrList += [_objs.GateString(
                ('Gz','Gx','Gx','Gx','Gx','Gz','Gz','Gx','Gx','Gx','Gx','Gz')*k+
                ('Gx',)*4, '(GzGxGxGxGxGzGzGxGxGxGxGz)^'+str(k)+'GxGxGxGx')]

        thetaSinStrList += [_objs.GateString(
                ('Gx','Gx','Gz','Gz')+
                ('Gz','Gx','Gx','Gx','Gx','Gz','Gz','Gx','Gx','Gx','Gx','Gz')*k+
                ('Gx',)*4,
                '(GxGxGzGz)(GzGxGxGxGxGzGzGxGxGxGxGz)^'+str(k)+'GxGxGxGx')]

        #From RPEToolsNewNew.py
        #thetaCosStrList += [_objs.GateString(
        #       ('Gz','Gx','Gx','Gx','Gx','Gz','Gz','Gx','Gx','Gx','Gx','Gz')*k,
        #       '(GzGxGxGxGxGzGzGxGxGxGxGz)^'+str(k))]
        #
        #thetaSinStrList += [_objs.GateString(
        #       ('Gx','Gx')+
        #       ('Gz','Gx','Gx','Gx','Gx','Gz','Gz','Gx','Gx','Gx','Gx','Gz')*k,
        #       'GxGx(GzGxGxGxGxGzGzGxGxGxGxGz)^'+str(k))]

    return thetaCosStrList, thetaSinStrList

def make_rpe_string_list_d(log2kMax):
    """
    Generates a dictionary that contains gate strings for all RPE cosine and
    sine experiments for all three angles.

    Parameters
    ----------
    log2kMax : int
       Maximum number of times to repeat an RPE "germ"

    Returns
    -------
    totalStrListD : dict
        A dictionary containing all gate strings for all sine and cosine
        experiments for alpha, epsilon, and theta.
        The keys of the returned dictionary are:

        - 'alpha','cos' : List of gate strings for cosine experiments used
          to determine alpha.
        - 'alpha','sin' : List of gate strings for sine experiments used to
          determine alpha.
        - 'epsilon','cos' : List of gate strings for cosine experiments used to
           determine epsilon.
        - 'epsilon','sin' : List of gate strings for sine experiments used to
          determine epsilon.
        - 'theta','cos' : List of gate strings for cosine experiments used to
          determine theta.
        - 'theta','sin' : List of gate strings for sine experiments used to
          determine theta.
        - 'totalStrList' : All above gate strings combined into one list;
          duplicates removed.
    """
    kList = [2**k for k in range(log2kMax+1)]
    alphaCosStrList, alphaSinStrList = make_rpe_alpha_str_lists_gx_gz(kList)
    epsilonCosStrList, epsilonSinStrList = make_rpe_epsilon_str_lists_gx_gz(kList)
    thetaCosStrList, thetaSinStrList = make_rpe_theta_str_lists_gx_gz(kList)
    totalStrList = alphaCosStrList + alphaSinStrList + epsilonCosStrList + epsilonSinStrList + thetaCosStrList + thetaSinStrList
    totalStrList = _tools.remove_duplicates(totalStrList) #probably superfluous

    stringListD = {}
    stringListD['alpha','cos'] = alphaCosStrList
    stringListD['alpha','sin'] = alphaSinStrList
    stringListD['epsilon','cos'] = epsilonCosStrList
    stringListD['epsilon','sin'] = epsilonSinStrList
    stringListD['theta','cos'] = thetaCosStrList
    stringListD['theta','sin'] = thetaSinStrList
    stringListD['totalStrList'] = totalStrList
    return stringListD

def make_rpe_data_set(gatesetOrDataset,stringListD,nSamples,sampleError='binomial',seed=None):
    """
    Generate a fake RPE DataSet using the probabilities obtained from a gateset.
    Is a thin wrapper for pygsti.construction.generate_fake_data, changing
    default behavior of sampleError, and taking a dictionary of gate strings
    as input.

    Parameters
    ----------
    gatesetOrDataset : GateSet or DataSet object
        If a GateSet, the gate set whose probabilities generate the data.
        If a DataSet, the data set whose frequencies generate the data.

    stringListD : Dictionary of list of (tuples or GateStrings)
        Each tuple or GateString contains gate labels and
        specifies a gate sequence whose counts are included
        in the returned DataSet.  The dictionary must have the key
        'totalStrList'; easiest if this dictionary is generated by
        make_rpe_string_list_d.

    nSamples : int or list of ints or None
        The simulated number of samples for each gate string.  This only
        has effect when  sampleError == "binomial" or "multinomial".  If
        an integer, all gate strings have this number of total samples. If
        a list, integer elements specify the number of samples for the
        corresponding gate string.  If None, then gatesetOrDataset must be
        a DataSet, and total counts are taken from it (on a per-gatestring
        basis).

    sampleError : string, optional
        What type of sample error is included in the counts.  Can be:

        - "none"  - no sampl error:
          counts are floating point numbers such that the exact
          probabilty can be found by the ratio of count / total.
        - "round" - same as "none", except counts are rounded to the nearest
          integer.
        - "binomial" - the number of counts is taken from a binomial
          distribution. Distribution has parameters p = probability of the
          gate string and n = number of samples.  This can only be used when
          there are exactly two SPAM labels in gatesetOrDataset.
        - "multinomial" - counts are taken from a multinomial distribution.
          Distribution has parameters p_k = probability of the gate string
          using the k-th SPAM label and n = number of samples.  This should not
          be used for RPE.

    seed : int, optional
        If not None, a seed for numpy's random number generator, which
        is used to sample from the binomial or multinomial distribution.

    Returns
    -------
    DataSet
       A static data set filled with counts for the specified gate strings.
    """
    return _dsc.generate_fake_data(gatesetOrDataset,
                                   stringListD['totalStrList'],
                                   nSamples,sampleError=sampleError,seed=seed)



#TODO savePlot arg is never used?
def rpe_ensemble_test(alphaTrue, epsilonTrue, Yrot, SPAMdepol, log2kMax, N, runs,
                  plot=False, savePlot=False):

    """ Experimental test function """
    kList = [2**k for k in range(log2kMax+1)]

    alphaCosStrList, alphaSinStrList = make_rpe_alpha_str_lists_gx_gz(kList)
    epsilonCosStrList, epsilonSinStrList = make_rpe_epsilon_str_lists_gx_gz(kList)
    thetaCosStrList, thetaSinStrList = make_rpe_theta_str_lists_gx_gz(kList)

    percentAlphaError = 100*_np.abs((_np.pi/2-alphaTrue)/alphaTrue)
    percentEpsilonError = 100*_np.abs((_np.pi/4 - epsilonTrue)/epsilonTrue)

    simGateset = _setc.build_gateset( [2], [('Q0',)],['Gi','Gx','Gz'],
                                      [ "I(Q0)", "X("+str(epsilonTrue)+",Q0)", "Z("+str(alphaTrue)+",Q0)"],
                                      prepLabels=["rho0"], prepExpressions=["0"],
                                      effectLabels=["E0"], effectExpressions=["1"],
                                      spamdefs={'plus': ('rho0','E0'), 'minus': ('rho0','remainder') } )


    gatesetAux1 = _setc.build_gateset( [2], [('Q0',)],['Gi','Gy','Gz'],
                                       [ "I(Q0)", "Y("+str(Yrot)+",Q0)", "Z(pi/2,Q0)"],
                                       prepLabels=["rho0"], prepExpressions=["0"],
                                       effectLabels=["E0"], effectExpressions=["1"],
                                       spamdefs={'plus': ('rho0','E0'), 'minus': ('rho0','remainder') } )

    simGateset.gates['Gx'] =  _objs.FullyParameterizedGate(
        _np.dot(_np.dot(_np.linalg.inv(gatesetAux1.gates['Gy']),simGateset.gates['Gx']),
                gatesetAux1.gates['Gy']))

    simGateset = simGateset.depolarize(spam_noise=SPAMdepol)

    thetaTrue = _tools.rpe.extract_theta(simGateset)

    SPAMerror = _np.dot(simGateset.effects['E0'].T,simGateset.preps['rho0'])[0,0]

    jMax = runs

    alphaHatListArray = _np.zeros([jMax,log2kMax+1],dtype='object')
    epsilonHatListArray = _np.zeros([jMax,log2kMax+1],dtype='object')
    thetaHatListArray = _np.zeros([jMax,log2kMax+1],dtype='object')

    alphaErrorArray = _np.zeros([jMax,log2kMax+1],dtype='object')
    epsilonErrorArray = _np.zeros([jMax,log2kMax+1],dtype='object')
    thetaErrorArray = _np.zeros([jMax,log2kMax+1],dtype='object')
    PhiFunErrorArray = _np.zeros([jMax,log2kMax+1],dtype='object')

    for j in range(jMax):
    #    simDS = _dsc.generate_fake_data(gateset3,alphaCosStrList+alphaSinStrList+epsilonCosStrList+epsilonSinStrList+thetaCosStrList+epsilonSinStrList,
        simDS = _dsc.generate_fake_data(
            simGateset, alphaCosStrList+alphaSinStrList+epsilonCosStrList+
            epsilonSinStrList+thetaCosStrList+thetaSinStrList,
            N,sampleError='binomial',seed=j)
        alphaErrorList = []
        epsilonErrorList = []
        thetaErrorList = []
        PhiFunErrorList = []
        alphaHatList = _tools.rpe.est_angle_list(simDS,alphaSinStrList,
                                                 alphaCosStrList,'alpha')
        epsilonHatList = _tools.rpe.est_angle_list(simDS,epsilonSinStrList,
                                                   epsilonCosStrList,'epsilon')
        thetaHatList,PhiFunList = _tools.rpe.est_theta_list(simDS,thetaSinStrList,
                                                 thetaCosStrList,epsilonHatList,
                                                 returnPhiFunList=True)
        for alphaTemp1 in alphaHatList:
            alphaErrorList.append(abs(alphaTrue - alphaTemp1))
        for epsilonTemp1 in epsilonHatList:
            epsilonErrorList.append(abs(epsilonTrue - epsilonTemp1))
    #        print abs(_np.pi/2-abs(alphaTemp1))
        for thetaTemp1 in thetaHatList:
            thetaErrorList.append(abs(thetaTrue - thetaTemp1))
        for PhiFunTemp1 in PhiFunList:
            PhiFunErrorList.append(PhiFunTemp1)

        alphaErrorArray[j,:] = _np.array(alphaErrorList)
        epsilonErrorArray[j,:] = _np.array(epsilonErrorList)
        thetaErrorArray[j,:] = _np.array(thetaErrorList)
        PhiFunErrorArray[j,:] = _np.array(PhiFunErrorList)

        alphaHatListArray[j,:] = _np.array(alphaHatList)
        epsilonHatListArray[j,:] = _np.array(epsilonHatList)
        thetaHatListArray[j,:] = _np.array(thetaHatList)

    #print "True alpha:",alphaTrue
    #print "True alpha:",alphaTrue
    #print "True alpha:",alphaTrue
    #print "True alpha:",alphaTrue
    #print "% true alpha deviation from target:", percentAlphaError

    if plot:
        import matplotlib as _mpl
        _mpl.pyplot.loglog(kList,_np.median(alphaErrorArray,axis=0),label='N='+str(N))

        _mpl.pyplot.loglog(kList,_np.array(kList)**-1.,'-o',label='1/k')
        _mpl.pyplot.xlabel('k')
        _mpl.pyplot.ylabel(r'$\alpha_z - \widehat{\alpha_z}$')
        _mpl.pyplot.title('RPE error in Z angle\n% error in Z angle '+str(percentAlphaError)+'%, % error in X angle '+str(percentEpsilonError)+'%\n% error in SPAM, '+str(100*SPAMerror)+'%, X-Z axis error '+str(Yrot)+'\nMedian of '+str(jMax)+' Trials')
        _mpl.pyplot.legend()

        _mpl.pyplot.show()

        _mpl.pyplot.loglog(kList,_np.median(epsilonErrorArray,axis=0),label='N='+str(N))

        _mpl.pyplot.loglog(kList,_np.array(kList)**-1.,'-o',label='1/k')
        _mpl.pyplot.xlabel('k')
        _mpl.pyplot.ylabel(r'$\epsilon_x - \widehat{\epsilon_x}$')
        _mpl.pyplot.title('RPE error in X angle\n% error in Z angle '+str(percentAlphaError)+'%, % error in X angle '+str(percentEpsilonError)+'%\n% error in SPAM, '+str(100*SPAMerror)+'%, X-Z axis error '+str(Yrot)+'\nMedian of '+str(jMax)+' Trials')
        _mpl.pyplot.legend()

        _mpl.pyplot.show()

        _mpl.pyplot.loglog(kList,_np.median(thetaErrorArray,axis=0),label='N='+str(N))

        _mpl.pyplot.loglog(kList,_np.array(kList)**-1.,'-o',label='1/k')
        _mpl.pyplot.xlabel('k')
        _mpl.pyplot.ylabel(r'$\theta_{xz} - \widehat{\theta_{xz}}$')
        _mpl.pyplot.title('RPE error in X axis angle\n% error in Z angle '+str(percentAlphaError)+'%, % error in X angle '+str(percentEpsilonError)+'%\n% error in SPAM, '+str(100*SPAMerror)+'%, X-Z axis error '+str(Yrot)+'\nMedian of '+str(jMax)+' Trials')
        _mpl.pyplot.legend()

        _mpl.pyplot.show()

        _mpl.pyplot.loglog(kList,_np.median(PhiFunErrorArray,axis=0),label='N='+str(N))

#        _mpl.pyplot.loglog(kList,_np.array(kList)**-1.,'-o',label='1/k')
        _mpl.pyplot.xlabel('k')
        _mpl.pyplot.ylabel(r'$\Phi func.$')
        _mpl.pyplot.title('RPE error in Phi func.\n% error in Z angle '+str(percentAlphaError)+'%, % error in X angle '+str(percentEpsilonError)+'%\n% error in SPAM, '+str(100*SPAMerror)+'%, X-Z axis error '+str(Yrot)+'\nMedian of '+str(jMax)+' Trials')
        _mpl.pyplot.legend()

    outputDict = {}
#    outputDict['alphaArray'] = alphaHatListArray
#    outputDict['alphaErrorArray'] = alphaErrorArray
#    outputDict['epsilonArray'] = epsilonHatListArray
#    outputDict['epsilonErrorArray'] = epsilonErrorArray
#    outputDict['thetaArray'] = thetaHatListArray
#    outputDict['thetaErrorArray'] = thetaErrorArray
#    outputDict['PhiFunErrorArray'] = PhiFunErrorArray
#    outputDict['alpha'] = alphaTrue
#    outputDict['epsilonTrue'] = epsilonTrue
#    outputDict['thetaTrue'] = thetaTrue
#    outputDict['Yrot'] = Yrot
#    outputDict['SPAMdepol'] = SPAMdepol#Input value to depolarize SPAM by
#    outputDict['SPAMerror'] = SPAMerror#<<E|rho>>
#    outputDict['gs'] = simGateset
#    outputDict['N'] = N

    return outputDict



#def make_rpe_data_set(inputGateset, log2kMax, N, seed = None, returnStringListDict = False):
#    """
#    Generate a fake RPE dataset.  At present, only works for kList of form [1,2,4,...,2**log2kMax]
#
#    Parameters
#    ----------
#    inputGateset : The gateset used to generate the data.
#    log2kMax : Maximum number of times to repeat an RPE "germ"
#    N : Number of clicks per experiment.
#    seed : Used to seed numpy's random number generator.  Default is None.
#    returnStringListDict : Do we want a dictionary of the sin and cos experiments for the various angles?  Default is False.
#
#    Returns
#    -------
#    simDS
#        The simulated dataset containing the RPE experiments.
#    stringListD
#        Dictionary of gate string lists for sin and cos experiments; is not returned by default.
#    """
#    kList = [2**k for k in range(log2kMax+1)]
#    alphaCosStrList, alphaSinStrList = make_alpha_str_lists_gx_gz(kList)
#    epsilonCosStrList, epsilonSinStrList = make_epsilon_str_lists_gx_gz(kList)
#    thetaCosStrList, thetaSinStrList = make_theta_str_lists_gx_gz(kList)
#    totalStrList = alphaCosStrList + alphaSinStrList + epsilonCosStrList + epsilonSinStrList + thetaCosStrList + thetaSinStrList
#    totalStrList = _tools.remove_duplicates(totalStrList)#This step is probably superfluous.
#    simDS = _dsc.generate_fake_data(inputGateset,totalStrList,N,sampleError='binomial',seed=seed)
#    if returnStringListDict:
#        stringListD = {}
#        stringListD['alpha','cos'] = alphaCosStrList
#        stringListD['alpha','sin'] = alphaSinStrList
#        stringListD['epsilon','cos'] = epsilonCosStrList
#        stringListD['epsilon','sin'] = epsilonSinStrList
#        stringListD['theta','cos'] = thetaCosStrList
#        stringListD['theta','sin'] = thetaSinStrList
#        return simDS, stringListD
#    else:
#        return simDS, None



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Classes for constructing confidence regions """

import numpy       as _np
import scipy.stats as _stats
import warnings    as _warnings
import itertools   as _itertools
from .. import optimize as _opt

from .gateset import P_RANK_TOL
from .verbosityprinter import VerbosityPrinter

# NON-MARKOVIAN ERROR BARS
#Connection with Robin's notes:
#
# Robins notes: pg 21 : want to set radius delta'(alpha,r2)
#   via: lambda(G) = lambda(G_mle) + delta'
#
# Connecting with pyGSTi Hessian (H) calculations:
#   lambda(G) = 2(maxLogL - logL(G)) ~ chi2_k   (as defined in notes)
#   lambda(G_mle) = 2(maxLogL - logL(G_mle))
#
#  expand logL around max:
#    logL(G_mle + dx) ~= logL(G_mle) - 1/2 dx*H*dx (no first order term)
#
#  Thus, delta'
#    delta' = lambda(G) - lambda(G_mle) = -2(log(G)-log(G_mle))
#           = dx*H*dx ==> delta' is just like C1 or Ck scaling factors
#                         used for computing normal confidence regions
#    (recall delta' is computed as the alpha-th quantile of a
#     non-central chi^2_{K',r2} where K' = #of gateset params and
#     r2 = lambda(G_mle) - (K-K'), where K = #max-model params (~#gatestrings)
#     is the difference between the expected (mean) lambda (=K-K') and what
#     we actually observe (=lambda(G_mle)).
#
#TODO: add "type" argument to ConfidenceRegion constructor?
#   ==> "std" or "non-markovian"

class ConfidenceRegion(object):
    """
    Encapsulates a hessian-based confidence region in gate-set space.

    A ConfidenceRegion computes and stores the quadratic form for an approximate
    confidence region based on a confidence level and a hessian, typically of either
    loglikelihood function or its proxy, the chi2 function.
    """

    def __init__(self, gateset, hessian, confidenceLevel,
                 hessianProjection="std", tol=1e-6, maxiter=10000,
                 nonMarkRadiusSq=0, linresponse_mlgst_params=None):
        """
        Initializes a new ConfidenceRegion.

        Parameters
        ----------
        gateset : GateSet
            the gate set point estimate that maximizes the logl or minimizes
            the chi2, and marks the point in gateset-space where the Hessian
            has been evaluated.

        hessian : numpy array
            A nParams x nParams Hessian matrix, where nParams is the number
            of dimensions of gateset space, i.e. gateset.num_params().  This 
            can and must be None when `hessianProjection` equals
            `linear response` (see below).

        confidenceLevel : float
            The confidence level as a percentage, i.e. between 0 and 100.

        hessianProjection : string, optional
            Specifies how (and whether) to project the given hessian matrix
            onto a non-gauge space.  Allowed values are:

            - 'std' -- standard projection onto the space perpendicular to the
              gauge space.
            - 'none' -- no projection is performed.  Useful if the supplied
              hessian has already been projected.
            - 'optimal gate CIs' -- a lengthier projection process in which a
              numerical optimization is performed to find the non-gauge space
              which minimizes the (average) size of the confidence intervals
              corresponding to gate (as opposed to SPAM vector) parameters.
            - 'intrinsic error' -- compute separately the intrinsic error
              in the gate and spam GateSet parameters and set weighting metric
              based on their ratio.
            - 'linear response' -- obtain elements of the Hessian via the
              linear response of a "forcing term".  This requres a likelihood
              optimization for *every* computed error bar, but avoids pre-
              computation of the entire Hessian matrix, which can be 
              prohibitively costly on large parameter spaces.

        tol : float, optional
            Tolerance for optimal Hessian projection.  Only used when
            hessianProjection == 'optimal gate CIs'

        maxiter : int, optional
            Maximum iterations for optimal Hessian projection.  Only used when
            hessianProjection == 'optimal gate CIs'

        nonMarkRadiusSq : float, optional
            When non-zero, "a non-Markovian error region" is constructed using
            this value as the squared "non-markovian radius". This specifies the
            portion of 2*(max-log-likelihood - gateset-log-likelihood) that we
            attribute to non-Markovian errors (typically the previous
            difference minus it's expected value, the difference in number of
            parameters between the maximal and gateset models).  If set to
            zero (the default), a standard and thereby statistically rigorous
            conficence region is created.  Non-zero values should only be
            supplied if you really know what you're doing.

        linresponse_mlgst_params : dict, optional
            Only used when `hessianProjection == 'linear response'`, this 
            dictionary gives the arguments passed to the :func:`do_mlgst`
            calls used to compute linear-response error bars.
        """

        assert(not(hessian is None and linresponse_mlgst_params is None)), \
            "Must supply either a Hessian matrix or MLGST parameters"

        proj_non_gauge = gateset.get_nongauge_projector()
        self.nNonGaugeParams = _np.linalg.matrix_rank(proj_non_gauge, P_RANK_TOL)
        self.nGaugeParams = gateset.num_params() - self.nNonGaugeParams

        #Project Hessian onto non-gauge space
        if hessian is not None:
            if hessianProjection == 'none':
                projected_hessian = hessian
            elif hessianProjection == 'std':
                projected_hessian = _np.dot(proj_non_gauge, _np.dot(hessian, proj_non_gauge))
            elif hessianProjection == 'optimal gate CIs':
                projected_hessian = _optProjectionForGateCIs(gateset, hessian, self.nNonGaugeParams,
                                                             self.nGaugeParams, confidenceLevel,
                                                             "L-BFGS-B", maxiter, maxiter,
                                                             tol, verbosity=3) #verbosity for DEBUG
            elif hessianProjection == 'intrinsic error':
                projected_hessian = _optProjectionFromSplit(gateset, hessian, confidenceLevel,
                                                            verbosity=3) #verbosity for DEBUG
            elif hessianProjection == 'linear response':
                raise ValueError("'hessian' must be None when using the " +
                                 "'linear response' hessian projection type")
            else:
                raise ValueError("Invalid value of hessianProjection argument: %s" % hessianProjection)
        else:
            assert(hessianProjection == 'linear response')
            projected_hessian = None

        #Scale projected Hessian for desired confidence level => quadratic form for confidence region
        # assume hessian gives Fisher info, so asymptotically normal => confidence interval = +/- seScaleFctr * 1/sqrt(hessian)
        # where seScaleFctr gives the scaling factor for a normal distribution, i.e. integrating the
        # std normal distribution between -seScaleFctr and seScaleFctr == confidenceLevel/100 (as a percentage)
        assert(confidenceLevel > 0.0 and confidenceLevel < 100.0)
        if confidenceLevel < 1.0:
            _warnings.warn("You've specified a %f%% confidence interval, " % confidenceLevel \
                               + "which is usually small.  Be sure to specify this" \
                               + "number as a percentage in (0,100) and not a fraction in (0,1)." )

        # Get constants C such that xT*Hessian*x = C gives contour for the desired confidence region.
        #  C1 == Single DOF case: constant for a single-DOF likelihood, (or a profile likelihood in our case)
        #  Ck == Total DOF case: constant for a region of the likelihood as a function of *all non-gauge* gateset parameters
        if nonMarkRadiusSq == 0.0: #use == to test for *exact* zero floating pt value as herald
            C1 = _stats.chi2.ppf(confidenceLevel/100.0, 1)
            Ck = _stats.chi2.ppf(confidenceLevel/100.0, self.nNonGaugeParams)

              # Alt. method to get C1: square the result of a single gaussian (normal distribution)
              #Note: scipy's ppf gives inverse of cdf, so want to know where cdf == the leftover probability on left side
            seScaleFctr = -_stats.norm.ppf((1.0 - confidenceLevel/100.0)/2.0) #std error scaling factor for desired confidence region
            assert(_np.isclose(C1, seScaleFctr**2))

            # save quadratic form Q s.t. xT*Q*x = 1 gives confidence region using C1, i.e. a
            #  region appropriate for generating 1-D confidence intervals.
            if projected_hessian is not None:
                self.regionQuadcForm = projected_hessian / C1
            else:
                self.regionQuadcForm = None

            self.intervalScaling = _np.sqrt( Ck / C1 ) # multiplicative scaling required to convert intervals
                                                   # to those obtained using a full (using Ck) confidence region.
            self.stdIntervalScaling = 1.0 # multiplicative scaling required to convert intervals
                                          # to *standard* (e.g. not non-Mark.) intervals.
            self.stdRegionScaling = self.intervalScaling # multiplicative scaling required to convert intervals
                                                  # to those obtained using a full *standard* confidence region.

        else:
            C1 = _stats.ncx2.ppf(confidenceLevel/100.0, 1, nonMarkRadiusSq)
            Ck = _stats.ncx2.ppf(confidenceLevel/100.0, self.nNonGaugeParams, nonMarkRadiusSq)

            # save quadratic form Q s.t. xT*Q*x = 1 gives confidence region using C1, i.e. a
            #  region appropriate for generating 1-D confidence intervals.
            if projected_hessian is not None:
                self.regionQuadcForm = projected_hessian / C1
                self.regionQuadcForm *=  _np.sqrt(self.nNonGaugeParams) #make a *worst case* non-mark. region...
            else:
                self.regionQuadcForm = None

            self.intervalScaling = _np.sqrt( Ck / C1 ) # multiplicative scaling required to convert intervals
                                                   # to those obtained using a full (using Ck) confidence region.

            stdC1 = _stats.chi2.ppf(confidenceLevel/100.0, 1)
            stdCk = _stats.chi2.ppf(confidenceLevel/100.0, self.nNonGaugeParams)
            self.stdIntervalScaling = _np.sqrt( stdC1 / C1 ) # see above description
            self.stdRegionScaling = _np.sqrt( stdCk / C1 ) # see above description

            _warnings.warn("Non-Markovian error bars are experimental and" +
                           " cannot be interpreted as standard error bars." +
                           " Proceed with caution!")

        #print "DEBUG: C1 =",C1," Ck =",Ck," scaling =",self.intervalScaling

        if self.regionQuadcForm is not None:
            #Invert *non-gauge-part* of quadratic by eigen-decomposing ->
            #   inverting the non-gauge eigenvalues -> re-constructing via eigenvectors.
            # (Note that Hessian & quadc form mxs are symmetric so eigenvalues == singular values)
            evals,U = _np.linalg.eigh(self.regionQuadcForm)  # regionQuadcForm = U * diag(evals) * U.dag (b/c U.dag == inv(U) )
            Udag = _np.conjugate(_np.transpose(U))
    
              #invert only the non-gauge eigenvalues (those with ordering index >= nGaugeParams)
            orderInds = [ el[0] for el in sorted( enumerate(evals), key=lambda x: abs(x[1])) ] # ordering index of each eigenvalue
            invEvals = _np.zeros( evals.shape, evals.dtype )
            for i in orderInds[self.nGaugeParams:]:
                invEvals[i] = 1.0/evals[i]
            #print "nGaugeParams = ",self.nGaugeParams; print invEvals #DEBUG
    
              #re-construct "inverted" quadratic form
            invDiagQuadcForm = _np.diag( invEvals )
            self.invRegionQuadcForm = _np.dot(U, _np.dot(invDiagQuadcForm, Udag))
            self.U, self.Udag = U, Udag
        else:
            self.invRegionQuadcForm = None
            self.U = self.Udag = None

        #Store params and offsets of gateset for future use
        self.gateset_offsets = gateset.get_vector_offsets()

        #Store list of profile-likelihood confidence intervals
        #  which == sqrt(diagonal els) of invRegionQuadcForm
        if self.invRegionQuadcForm is not None:
            self.profLCI = [ _np.sqrt(abs(self.invRegionQuadcForm[k,k])) for k in range(len(evals))]
            self.profLCI = _np.array( self.profLCI, 'd' )
        else:
            self.profLCI = None

        self.gateset = gateset
        self.level = confidenceLevel #a percentage, i.e. btwn 0 and 100

        self.mlgst_params = linresponse_mlgst_params
        self._C1 = C1 #save for linear response scaling
        self.mlgst_evaltree_cache = {} #for _do_mlgst_base speedup

        #DEBUG
        #print "DEBUG HESSIAN:"
        #print "shape = ",hessian.shape
        #print "nGaugeParams = ",self.nGaugeParams
        #print "nNonGaugeParams = ",self.nNonGaugeParams
        ##print "Eval,InvEval = ",zip(evals,invEvals)
        #N = self.regionQuadcForm.shape[0]
        ##diagEls = [ self.regionQuadcForm[k,k] for k in range(N) ]
        #invdiagEls = sorted([ abs(self.invRegionQuadcForm[k,k]) for k in range(N) ])
        #evals = sorted( _np.abs(_np.linalg.eigvals( self.regionQuadcForm )) )
        #invEvals = sorted( _np.abs(_np.linalg.eigvals( self.invRegionQuadcForm )) )
        #print "index  sorted(abs(inv-diag))     sorted(eigenval)    sorted(inv-eigenval):"
        #for i,(invDiag,eigenval,invev) in enumerate(zip(invdiagEls,evals,invEvals)):
        #    print "%d %15g %15g %15g" % (i,invDiag,eigenval,invev)
        #
        #import sys
        #sys.exit()

    def has_hessian(self):
        """
        Returns whether or not the full Hessian has already been computed.

        When True, computation of confidence regions and intervals is
        fast, since the difficult work of computing the inverse Hessian is
        already done.  When False, a slower method must be used to estimate
        the necessary portion of the Hessian.  The result of this function
        is often used to decide whether or not to proceed with an error-bar
        computation.

        Returns
        -------
        bool
        """
        return bool(self.invRegionQuadcForm is not None)


    def get_gateset(self):
        """
        Retrieve the associated gate set.

        Returns
        -------
        GateSet
            the gate set marking the center location of this confidence region.
        """
        return self.gateset


    def get_profile_likelihood_confidence_intervals(self, label=None):
        """
        Retrieve the profile-likelihood confidence intervals for a specified
        gate set object (or all such intervals).

        Parameters
        ----------
        label : string, optional
            If not None, can be either a gate or SPAM vector label
            to specify the confidence intervals corresponding to gate, rhoVec,
            or EVec parameters respectively.  If None, then intervals corresponding
            to all of the gateset's parameters are returned.

        Returns
        -------
        numpy array
            One-dimensional array of (positive) interval half-widths which specify
            a symmetric confidence interval.
        """
        if label is None:
            return self.profLCI
        else:
            start,end = self.gateset_offsets[label]
            return self.profLCI[start:end]

    def get_gate_fn_confidence_interval(self, fnOfGate, gateLabel, eps=1e-7,
                                        returnFnVal=False, verbosity=0):
        """
        Compute the confidence interval for a function of a single gate.

        Parameters
        ----------
        fnOfGate : function
            A function which takes as its only argument a gate matrix.  The
            returned confidence interval is based on linearizing this function
            and propagating the gateset-space confidence region.

        gateLabel : string
            The label specifying which gate to use in evaluations of fnOfGate.

        eps : float, optional
            Step size used when taking finite-difference derivatives of fnOfGate.

        returnFnVal : bool, optional
            If True, return the value of fnOfGate along with it's confidence
            region half-widths.

        verbosity : int, optional
            Specifies level of detail in standard output.

        Returns
        -------
        df : float or numpy array
            Half-widths of confidence intervals for each of the elements
            in the float or array returned by fnOfGate.  Thus, shape of
            df matches that returned by fnOfGate.

        f0 : float or numpy array
            Only returned when returnFnVal == True. Value of fnOfGate
            at the gate specified by gateLabel.
        """

        nParams = self.gateset.num_params()

        gateObj = self.gateset.gates[gateLabel].copy() # copy because we add eps to this gate
        gateMx = _np.asarray(gateObj).copy()
        gpo = self.gateset_offsets[gateLabel][0] #starting "gate parameter offset"

        f0 = fnOfGate(gateMx) #function value at "base point" gateMx
        nGateParams = gateObj.num_params()
        gateVec0 = gateObj.to_vector()

        #Get finite difference derivative gradF that is shape (nParams, <shape of f0>)
        if type(f0) == float:
            gradF = _np.zeros( nParams, 'd' ) #assume all functions are real-valued for now...
        else:
            gradSize = (nParams,) + tuple(f0.shape)
            gradF = _np.zeros( gradSize, f0.dtype ) #assume all functions are real-valued for now...

        for i in range(nGateParams):
            gateVec = gateVec0.copy(); gateVec[i] += eps; gateObj.from_vector(gateVec)
            gradF[gpo + i] = ( fnOfGate( gateObj ) - f0 ) / eps

        return self._compute_df_from_gradF(gradF, f0, returnFnVal, verbosity)


    def get_prep_fn_confidence_interval(self, fnOfPrep, prepLabel, eps=1e-7,
                                        returnFnVal=False, verbosity=0):
        """
        Compute the confidence interval for a function of a single state prep.

        Parameters
        ----------
        fnOfPrep : function
            A function which takes as its only argument a prepration vector.  The
            returned confidence interval is based on linearizing this function
            and propagating the gateset-space confidence region.

        prepLabel : string
            The label specifying which preparation to use in evaluations of fnOfPrep.

        eps : float, optional
            Step size used when taking finite-difference derivatives of fnOfPrep.

        returnFnVal : bool, optional
            If True, return the value of fnOfPrep along with it's confidence
            region half-widths.

        verbosity : int, optional
            Specifies level of detail in standard output.

        Returns
        -------
        df : float or numpy array
            Half-widths of confidence intervals for each of the elements
            in the float or array returned by fnOfPrep.  Thus, shape of
            df matches that returned by fnOfPrep.

        f0 : float or numpy array
            Only returned when returnFnVal == True. Value of fnOfPrep
            at the state preparation specified by prepLabel.
        """

        nParams = self.gateset.num_params()

        prepObj = self.gateset.preps[prepLabel].copy() # copy because we add eps to this gate
        spamVec = _np.asarray(prepObj).copy()
        gpo = self.gateset_offsets[prepLabel][0] #starting "gateset parameter offset"

        f0 = fnOfPrep(spamVec) #function value at "base point"
        nPrepParams = prepObj.num_params()
        prepVec0 = prepObj.to_vector()

        #Get finite difference derivative gradF that is shape (nParams, <shape of f0>)
        if type(f0) == float:
            gradF = _np.zeros( nParams, 'd' ) #assume all functions are real-valued for now...
        else:
            gradSize = (nParams,) + tuple(f0.shape)
            gradF = _np.zeros( gradSize, f0.dtype ) #assume all functions are real-valued for now...

        for i in range(nPrepParams):
            prepVec = prepVec0.copy(); prepVec[i] += eps; prepObj.from_vector(prepVec)
            gradF[gpo + i] = ( fnOfPrep( prepObj ) - f0 ) / eps

        return self._compute_df_from_gradF(gradF, f0, returnFnVal, verbosity)


    def get_effect_fn_confidence_interval(self, fnOfEffect, effectLabel, eps=1e-7,
                                        returnFnVal=False, verbosity=0):
        """
        Compute the confidence interval for a function of a single POVM effect.

        Parameters
        ----------
        fnOfEffect : function
            A function which takes as its only argument a POVM vector.  The
            returned confidence interval is based on linearizing this function
            and propagating the gateset-space confidence region.

        effectLabel : string
            The label specifying which POVM to use in evaluations of fnOfEffect.

        eps : float, optional
            Step size used when taking finite-difference derivatives of fnOfEffect.

        returnFnVal : bool, optional
            If True, return the value of fnOfEffect along with it's confidence
            region half-widths.

        verbosity : int, optional
            Specifies level of detail in standard output.

        Returns
        -------
        df : float or numpy array
            Half-widths of confidence intervals for each of the elements
            in the float or array returned by fnOfEffect.  Thus, shape of
            df matches that returned by fnOfEffect.

        f0 : float or numpy array
            Only returned when returnFnVal == True. Value of fnOfEffect
            at the POVM effect specified by effectLabel.
        """

        nParams = self.gateset.num_params()

        effectObj = self.gateset.effects[effectLabel].copy() # copy because we add eps to this gate
        spamVec = _np.asarray(effectObj).copy()
        f0 = fnOfEffect(spamVec) #function value at "base point"

        if effectLabel not in self.gateset_offsets: #e.g. "remainder" is not...
            #Assume this effect label has not official "parameters" and just
            # return 0 as the confidence interval.
            return (0.0,f0) if returnFnVal else 0.0

        gpo = self.gateset_offsets[effectLabel][0] #starting "gateset parameter offset"
        nEffectParams = effectObj.num_params()
        effectVec0 = effectObj.to_vector()

        #Get finite difference derivative gradF that is shape (nParams, <shape of f0>)
        if type(f0) == float:
            gradF = _np.zeros( nParams, 'd' ) #assume all functions are real-valued for now...
        else:
            gradSize = (nParams,) + tuple(f0.shape)
            gradF = _np.zeros( gradSize, f0.dtype ) #assume all functions are real-valued for now...

        for i in range(nEffectParams):
            effectVec = effectVec0.copy(); effectVec[i] += eps; effectObj.from_vector(effectVec)
            gradF[gpo + i] = ( fnOfEffect( effectObj ) - f0 ) / eps

        return self._compute_df_from_gradF(gradF, f0, returnFnVal, verbosity)


    def get_gateset_fn_confidence_interval(self, fnOfGateset, eps=1e-7, returnFnVal=False, verbosity=0):
        """
        Compute the confidence interval for a function of a GateSet.

        Parameters
        ----------
        fnOfGateset : function
            A function which takes as its only argument a GateSet object.  The
            returned confidence interval is based on linearizing this function
            and propagating the gateset-space confidence region.

        eps : float, optional
            Step size used when taking finite-difference derivatives of fnOfGateset.

        returnFnVal : bool, optional
            If True, return the value of fnOfGateset along with it's confidence
            region half-widths.

        verbosity : int, optional
            Specifies level of detail in standard output.

        Returns
        -------
        df : float or numpy array
            Half-widths of confidence intervals for each of the elements
            in the float or array returned by fnOfGateset.  Thus, shape of
            df matches that returned by fnOfGateset.

        f0 : float or numpy array
            Only returned when returnFnVal == True. Value of fnOfGateset
            at the gate specified by gateLabel.
        """

        nParams = self.gateset.num_params()

        gateset = self.gateset.copy() # copy because we add eps to this gateset

        f0 = fnOfGateset(gateset) #function value at "base point" gateMx
        gatesetVec0 = gateset.to_vector()
        assert(len(gatesetVec0) == nParams)

        #Get finite difference derivative gradF that is shape (nParams, <shape of f0>)
        if type(f0) == float:
            gradF = _np.zeros( nParams, 'd' ) #assume all functions are real-valued for now...
        else:
            gradSize = (nParams,) + tuple(f0.shape)
            gradF = _np.zeros( gradSize, f0.dtype ) #assume all functions are real-valued for now...

        for i in range(nParams):
            gatesetVec = gatesetVec0.copy(); gatesetVec[i] += eps
            gateset.from_vector(gatesetVec)
            gradF[i] = ( fnOfGateset(gateset) - f0 ) / eps

        return self._compute_df_from_gradF(gradF, f0, returnFnVal, verbosity)


    def get_spam_fn_confidence_interval(self, fnOfSpamVecs, eps=1e-7, returnFnVal=False, verbosity=0):
        """
        Compute the confidence interval for a function of spam vectors.

        Parameters
        ----------
        fnOfSpamVecs : function
            A function which takes two arguments, rhoVecs and EVecs, each of which
            is a list of column vectors.  Note that the EVecs list will include
            *all* the effect vectors, including the a "compliment" vector.  The
            returned confidence interval is based on linearizing this function
            and propagating the gateset-space confidence region.

        eps : float, optional
            Step size used when taking finite-difference derivatives of fnOfSpamVecs.

        returnFnVal : bool, optional
            If True, return the value of fnOfSpamVecs along with it's confidence
            region half-widths.

        verbosity : int, optional
            Specifies level of detail in standard output.

        Returns
        -------
        df : float or numpy array
            Half-widths of confidence intervals for each of the elements
            in the float or array returned by fnOfSpamVecs.  Thus, shape of
            df matches that returned by fnOfSpamVecs.

        f0 : float or numpy array
            Only returned when returnFnVal == True. Value of fnOfSpamVecs.
        """
        nParams = self.gateset.num_params()

        f0 = fnOfSpamVecs(self.gateset.get_preps(), self.gateset.get_effects())
          #Note: .get_Evecs() can be different from .EVecs b/c the former includes compliment EVec

        #Get finite difference derivative gradF that is shape (nParams, <shape of f0>)
        if type(f0) == float:
            gradF = _np.zeros( nParams, 'd' ) #assume all functions are real-valued for now...
        else:
            gradSize = (nParams,) + tuple(f0.shape)
            gradF = _np.zeros( gradSize, f0.dtype ) #assume all functions are real-valued for now...


        gsEps = self.gateset.copy()

        #loop just over parameterized objects - don't use get_preps() here...
        for prepLabel,rhoVec in self.gateset.preps.items():
            nRhoParams = rhoVec.num_params()
            off = self.gateset_offsets[prepLabel][0]
            vec = rhoVec.to_vector()
            for i in range(nRhoParams):
                vecEps = vec.copy(); vecEps[i] += eps
                gsEps.preps[prepLabel].from_vector(vecEps) #update gsEps parameters
                gradF[off + i] = ( fnOfSpamVecs( gsEps.get_preps(),
                                   gsEps.get_effects() ) - f0 ) / eps
            gsEps.preps[prepLabel] = rhoVec.copy()  #reset gsEps (copy() just to be safe)

        #loop just over parameterized objects - don't use get_effects() here...
        for ELabel,EVec in self.gateset.effects.items():
            nEParams = EVec.num_params()
            off = self.gateset_offsets[ELabel][0]
            vec = EVec.to_vector()
            for i in range(nEParams):
                vecEps = vec.copy(); vecEps[i] += eps
                gsEps.effects[ELabel].from_vector(vecEps) #update gsEps parameters
                gradF[off + i] = ( fnOfSpamVecs( gsEps.get_preps(),
                                   gsEps.get_effects() ) - f0 ) / eps
            gsEps.effects[ELabel] = EVec.copy()  #reset gsEps (copy() just to be safe)

        return self._compute_df_from_gradF(gradF, f0, returnFnVal, verbosity)


    def _compute_df_from_gradF(self, gradF, f0, returnFnVal, verbosity):
        if self.regionQuadcForm is None:
            df = self._compute_df_from_gradF_linresponse(
                gradF, f0, verbosity)
        else:
            df = self._compute_df_from_gradF_hessian(
                gradF, f0, verbosity)
        return (df, f0) if returnFnVal else df


    def _compute_df_from_gradF_linresponse(self, gradF, f0, verbosity):
        from .. import algorithms as _alg
        assert(self.mlgst_params is not None)

        if hasattr(f0,'dtype') and f0.dtype == _np.dtype("complex"):
            raise NotImplementedError("Can't handle complex-valued functions yet")

        #massage gradF, which has shape (nParams,) + f0.shape
        # to that expected by _do_mlgst_base, which is
        # (flat_f0_size, nParams)
        if len(gradF.shape) == 1:
            gradF.shape = (1,gradF.shape[0])
        else:
            flatDim = _np.prod(f0.shape)
            gradF.shape = (gradF.shape[0], flatDim)
            gradF = _np.transpose(gradF) #now shape == (flatDim, nParams)
        assert(len(gradF.shape) == 2)

        mlgst_args = self.mlgst_params.copy()
        mlgst_args['startGateset'] = self.gateset
        mlgst_args['forcefn_grad'] = gradF
        mlgst_args['shiftFctr'] = 100.0
        mlgst_args['evaltree_cache'] = self.mlgst_evaltree_cache
        maxLogL, bestGS = _alg.core._do_mlgst_base(**mlgst_args)
        bestGS = _alg.gaugeopt_to_target(bestGS, self.gateset) #maybe more params here?
        norms = _np.array([_np.dot(gradF[i],gradF[i]) for i in range(gradF.shape[0])])
        delta2 = _np.abs(_np.dot(gradF, bestGS.to_vector() - self.gateset.to_vector()) \
            * _np.where(norms > 1e-10, 1.0/norms, 0.0))
        delta2 *= self._C1 #scaling appropriate for confidence level
        delta = _np.sqrt(delta2) # error^2 -> error

        if hasattr(f0,'shape'):
            delta.shape = f0.shape #reshape to un-flattened
        else:
            assert(type(f0) == float)
            delta = float(delta)

        return delta
        


    def _compute_df_from_gradF_hessian(self, gradF, f0, verbosity):
        """
        Internal function which computes error bars given an function value
        and gradient (using linear approx. to function)
        """

        #Compute df = sqrt( gradFu.dag * 1/D * gradFu )
        #  where regionQuadcForm = U * D * U.dag and gradFu = U.dag * gradF
        #  so df = sqrt( gradF.dag * U * 1/D * U.dag * gradF )
        #        = sqrt( gradF.dag * invRegionQuadcForm * gradF )

        printer = VerbosityPrinter.build_printer(verbosity)

        printer.log("gradF = %s" % gradF)

        if type(f0) == float:
            gradFdag = _np.conjugate(_np.transpose(gradF))

            #DEBUG
            #arg = _np.dot(gradFdag, _np.dot(self.invRegionQuadcForm, gradF))
            #print "HERE: taking sqrt(abs(%s))" % arg

            df = _np.sqrt( abs(_np.dot(gradFdag, _np.dot(self.invRegionQuadcForm, gradF))) )
        else:
            fDims = len(f0.shape)
            gradF = _np.rollaxis(gradF, 0, 1+fDims) # roll parameter axis to be the last index, preceded by f-shape
            df = _np.empty( f0.shape, f0.dtype)

            if f0.dtype == _np.dtype("complex"): #real and imaginary parts separately
                if fDims == 0: #same as float case above
                    gradFdag = _np.transpose(gradF)
                    df = _np.sqrt( abs( _np.dot(gradFdag.real, _np.dot(self.invRegionQuadcForm, gradF.real))) ) \
                        + 1j * _np.sqrt( abs( _np.dot(gradFdag.imag, _np.dot(self.invRegionQuadcForm, gradF.imag))) )
                elif fDims == 1:
                    for i in range(f0.shape[0]):
                        gradFdag = _np.transpose(gradF[i])
                        df[i] = _np.sqrt( abs( _np.dot(gradFdag.real, _np.dot(self.invRegionQuadcForm, gradF[i].real))) ) \
                            + 1j * _np.sqrt( abs( _np.dot(gradFdag.imag, _np.dot(self.invRegionQuadcForm, gradF[i].imag))) )
                elif fDims == 2:
                    for i in range(f0.shape[0]):
                        for j in range(f0.shape[1]):
                            gradFdag = _np.transpose(gradF[i,j])
                            df[i,j] = _np.sqrt( abs( _np.dot(gradFdag.real, _np.dot(self.invRegionQuadcForm, gradF[i,j].real))) ) \
                                + 1j * _np.sqrt( abs( _np.dot(gradFdag.imag, _np.dot(self.invRegionQuadcForm, gradF[i,j].imag))) )
                else:
                    raise ValueError("Unsupported number of dimensions returned by fnOfGate or fnOfGateset: %d" % fDims)

            else: #assume real -- so really don't need conjugate calls below
                if fDims == 0: #same as float case above
                    gradFdag = _np.conjugate(_np.transpose(gradF))

                    #DEBUG
                    #arg = _np.dot(gradFdag, _np.dot(self.invRegionQuadcForm, gradF))
                    #print "HERE2: taking sqrt(abs(%s))" % arg

                    df = _np.sqrt( abs( _np.dot(gradFdag, _np.dot(self.invRegionQuadcForm, gradF))) )
                elif fDims == 1:
                    for i in range(f0.shape[0]):
                        gradFdag = _np.conjugate(_np.transpose(gradF[i]))
                        df[i] = _np.sqrt( abs(_np.dot(gradFdag, _np.dot(self.invRegionQuadcForm, gradF[i]))) )
                elif fDims == 2:
                    for i in range(f0.shape[0]):
                        for j in range(f0.shape[1]):
                            gradFdag = _np.conjugate(_np.transpose(gradF[i,j]))
                            df[i,j] = _np.sqrt( abs(_np.dot(gradFdag, _np.dot(self.invRegionQuadcForm, gradF[i,j]))) )
                else:
                    raise ValueError("Unsupported number of dimensions returned by fnOfGate or fnOfGateset: %d" % fDims)

        printer.log("df = %s" % df)

        return df

    def __getstate__(self):
        #Return the state (for pickling) -- *don't* pickle any Comm objects
        to_pickle = self.__dict__.copy()
        if self.mlgst_params and self.mlgst_params.has_key("comm"):
            del self.mlgst_params['comm'] # one *cannot* pickle Comm objects
        return to_pickle



def _optProjectionForGateCIs(gateset, base_hessian, nNonGaugeParams, nGaugeParams,
                             level, method = "L-BFGS-B", maxiter = 10000,
                             maxfev = 10000, tol = 1e-6, verbosity = 0):
    printer = VerbosityPrinter.build_printer(verbosity)

    printer.log('', 3)
    printer.log("--- Hessian Projector Optimization for gate CIs (%s) ---" % method, 2, indentOffset=-1)

    def objective_func(vectorM):
        matM = vectorM.reshape( (nNonGaugeParams,nGaugeParams) )
        proj_extra = gateset.get_nongauge_projector(nonGaugeMixMx=matM)
        projected_hessian_ex = _np.dot(proj_extra, _np.dot(base_hessian, proj_extra))

        ci = ConfidenceRegion(gateset, projected_hessian_ex, level, hessianProjection="none")
        gateCIs = _np.concatenate( [ ci.get_profile_likelihood_confidence_intervals(gl).flatten()
                                     for gl in gateset.gates] )
        return _np.sqrt( _np.sum(gateCIs**2) )

    #Run Minimization Algorithm
    startM = _np.zeros( (nNonGaugeParams,nGaugeParams), 'd')
    x0 = startM.flatten()
    print_obj_func = _opt.create_obj_func_printer(objective_func)
    minSol = _opt.minimize(objective_func, x0,
                                    method=method, maxiter=maxiter,
                                    maxfev=maxfev, tol=tol,
                                    callback = print_obj_func if verbosity > 2 else None)

    mixMx = minSol.x.reshape( (nNonGaugeParams,nGaugeParams) )
    proj_extra = gateset.get_nongauge_projector(nonGaugeMixMx=mixMx)
    projected_hessian_ex = _np.dot(proj_extra, _np.dot(base_hessian, proj_extra))

    printer.log('The resulting min sqrt(sum(gateCIs**2)): %g' % minSol.fun, 2)

    return projected_hessian_ex


def _optProjectionFromSplit(gateset, base_hessian, level, verbosity = 0):
    printer = VerbosityPrinter.build_printer(verbosity)

    printer.log('', 3)
    printer.log("--- Hessian Projector Optimization from separate SPAM and Gate weighting ---", 2, indentOffset=-1)


    #get gate-intrinsic-error
    proj = gateset.get_nongauge_projector(itemWeights={'gates':1.0,'spam': 0.0})
    projected_hessian = _np.dot(proj, _np.dot(base_hessian, proj))
    ci = ConfidenceRegion(gateset, projected_hessian, level, hessianProjection="none")
    gateCIs = _np.concatenate( [ ci.get_profile_likelihood_confidence_intervals(gl).flatten()
                                     for gl in gateset.gates] )
    gate_intrinsic_err = _np.sqrt( _np.sum(gateCIs**2) )

    #get spam-intrinsic-error
    proj = gateset.get_nongauge_projector(itemWeights={'gates':0.0,'spam': 1.0})
    projected_hessian = _np.dot(proj, _np.dot(base_hessian, proj))
    ci = ConfidenceRegion(gateset, projected_hessian, level, hessianProjection="none")
    spamCIs = _np.concatenate( [ ci.get_profile_likelihood_confidence_intervals(gl).flatten()
                                     for sl in _itertools.chain(iter(gateset.preps),
                                                                iter(gateset.effects))] )
    spam_intrinsic_err = _np.sqrt( _np.sum(spamCIs**2) )

    ratio = gate_intrinsic_err / spam_intrinsic_err
    proj = gateset.get_nongauge_projector(itemWeights={'gates':1.0,'spam': ratio})
    projected_hessian = _np.dot(proj, _np.dot(base_hessian, proj))

    if printer.verbosity >= 2:
        #Create ci here just to extract #'s for print stmts
        ci = ConfidenceRegion(gateset, projected_hessian, level, hessianProjection="none")
        gateCIs = _np.concatenate( [ ci.get_profile_likelihood_confidence_intervals(gl).flatten()
                                         for gl in gateset.gates] )
        spamCIs = _np.concatenate( [ ci.get_profile_likelihood_confidence_intervals(gl).flatten()
                                         for sl in _itertools.chain(iter(gateset.preps),
                                                                    iter(gateset.effects))] )
        gate_err = _np.sqrt( _np.sum(gateCIs**2) )
        spam_err = _np.sqrt( _np.sum(spamCIs**2) )
        printer.log('Resulting intrinsic errors: %g (gates), %g (spam)' %
                    (gate_intrinsic_err, spam_intrinsic_err), 2)
        printer.log('Resulting sqrt(sum(gateCIs**2)): %g' % gate_err, 2)
        printer.log('Resulting sqrt(sum(spamCIs**2)): %g' % spam_err, 2)

    return projected_hessian



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" A custom conjugate gradient descent algorithm """

import numpy as _np

try:
    from scipy.optimize import Result as _optResult #for earlier scipy versions
except:
    from scipy.optimize import OptimizeResult as _optResult #for later scipy versions


def fmax_cg(f, x0, maxiters=100, tol=1e-8, dfdx_and_bdflag = None, xopt=None):
    """
    Custom conjugate-gradient (CG) routine for maximizing a function.

    This function runs slower than scipy.optimize's 'CG' method, but doesn't
    give up or get stuck as easily, and so sometimes can be a better option.

    Parameters
    ----------
    fn : function
        The function to minimize.

    x0 : numpy array
        The starting point (argument to fn).

    maxiters : int, optional
        Maximum iterations.

    tol : float, optional
        Tolerace for convergence (compared to absolute difference in f)

    dfdx_and_bdflag : function, optional
        Function to compute jacobian of f as well as a boundary-flag.

    xopt : numpy array, optional
        Used for debugging, output can be printed relating current optimum
        relative xopt, assumed to be a known good optimum.

    Returns
    -------
    scipy.optimize.Result object
        Includes members 'x', 'fun', 'success', and 'message'.  **Note:** returns
        the negated maximum in 'fun' in order to conform to the return value of
        other minimization routines.
    """

    MIN_STEPSIZE = 1e-8
    FINITE_DIFF_STEP = 1e-4
    RESET = 5
    stepsize = 1e-6

    #if no dfdx specifed, use finite differences
    if dfdx_and_bdflag is None:
        dfdx_and_bdflag = lambda x: _finite_diff_dfdx_and_bdflag(f,x,FINITE_DIFF_STEP)

    step = 0
    x = x0; last_fx = f(x0); last_x = x0
    lastchange = 0
    lastgradnorm = 0.0 # Safer than relying on uninitialized variables
    lastgrad     = 0.0
    if last_fx is None: raise ValueError("fmax_cg was started out of bounds!")
    while( step < maxiters and ((stepsize>MIN_STEPSIZE) or (step%RESET !=1 and RESET>1))):

        grad,boundaryFlag = dfdx_and_bdflag(x)
        gradnorm = _np.dot(grad,grad)
        if step%RESET == 0: #reset change == gradient
            change = grad[:]
        else: # add gradient to change (conjugate gradient)
            #beta = gradnorm / lastgradnorm # Fletcher-Reeves
            beta = (gradnorm - _np.dot(grad,lastgrad))/lastgradnorm # Polak-Ribiere
            #beta = (gradnorm - _np.dot(grad,lastgrad))/(_np.dot(lastchange,grad-lastgrad)) #Hestenes-Stiefel
            change = grad + beta * lastchange

        for i in range(len(change)):
            if boundaryFlag[i]*change[i] > 0:
                change[i] = 0
                print("DEBUG: fmax Preventing motion along dim %s" % i)

        if max(abs(change)) == 0:
            print("Warning: Completely Boxed in!")
            fx = last_fx; x = last_x
            assert( abs(last_fx - f(last_x)) < 1e-6)
            break
            #i = list(abs(grad)).index(min(abs(grad)))
            #change[i] = -boundaryFlag[i] * 1.0 # could pick a random direction to move in?
            #gradnorm = 1.0  # punt...

        lastgrad = grad
        lastgradnorm = gradnorm

        multiplier = 1.0/max(abs(change))
        change *= multiplier

        # Now "change" has largest element 1.  Time to do a linear search to find optimal stepsize.
        # If the last step had crazy short length, reset stepsize
        if stepsize < MIN_STEPSIZE: stepsize = MIN_STEPSIZE;
        g = lambda s: f(x+s*change) # f along a line given by changedir.  Argument to function is stepsize.
        stepsize = _maximize1D(g,0,abs(stepsize),last_fx)  #find optimal stepsize along change direction

        predicted_difference = stepsize * _np.dot(grad,change)
        if xopt is not None: xopt_dot = _np.dot(change, xopt-x) / (_np.linalg.norm(change) * _np.linalg.norm(xopt-x))
        x += stepsize * change; fx = f(x)
        difference = fx - last_fx
        print("DEBUG: Max iter ", step, ": f=",fx,", dexpct=",predicted_difference-difference, \
            ", step=",stepsize,", xopt_dot=", xopt_dot if xopt is not None else "--", \
            ", chg_dot=",_np.dot(change,lastchange)/(_np.linalg.norm(change)*_np.linalg.norm(lastchange)+1e-6))

        if abs(difference) < tol: break #Convergence condition

        lastchange = change
        last_fx = fx
        last_x = x.copy()
        step += 1

    print("Finished Custom Contrained Newton CG Method")
    print(" iterations = %d" % step)
    print(" maximum f = %g" % fx)

    solution = _optResult()
    solution.x = x; solution.fun = -fx if fx is not None else None  # negate maximum to conform to other minimization routines
    if step < maxiters:
        solution.success = True
    else:
        solution.success = False
        solution.message = "Maximum iterations exceeded"
    return solution


# Minimize g(s), given (s1,g1=g(s1)) as a starting point and guess, s2 for minimum
def _maximize1D(g,s1,s2,g1):

    PHI = (1.0+_np.sqrt(5.0))/2 #golden ratio
    TOL = 1e-10; FRAC_TOL = 1e-6

    # Note (s1,g1) and s2 are given.  Start with bracket (s1,s2,s3)
    #s3 = s2*(1.0+PHI); g2 = g(s2); g3 = g(s3)
    s3 = s2 + PHI*(s2-s1); g2 = g(s2); g3 = g(s3)
    # s4,g4 = None,None
    s1_on_bd = s3_on_bd = False

    #print "DEBUG: BEGIN MAX 1D: s1,s3=", (s1,s3)

    assert(g1 is not None or g3 is not None)
    if g1 is None or g3 is None:
        if g1 is None: s1,g1 = _findBoundary(g,s3,s1); s1_on_bd = True
        if g3 is None: s3,g3 = _findBoundary(g,s1,s3); s3_on_bd = True
        s2 = s1+(s3-s1)/PHI; g2 = g(s2)


    while( (abs(s3-s1) > TOL) and (abs(s3-s1) > FRAC_TOL*(abs(s3)+abs(s1))) ):
        #print "DEBUG: Max1D iter: (",s1,",",g1,") (",s2,",",g2,") (",s3,",",g3,")"
        if g3>g2:
            if g2>=g1: # Expand to the right
                if s3_on_bd: print("** Returning on bd"); return s3 #can't expand any further to right
                s2,g2 = s3,g3
                s3 = s1 + (s3-s1)*PHI; g3 = g(s3)
                if g3 is None:
                    s3,g3 = _findBoundary(g,s2,s3)
                    s3_on_bd = True
            else:  # contract to the left.
                s3,g3 = s2,g2
                s2 = s1 + (s3-s1)/PHI; g2 = g(s2)
        else:
            if g2<=g1: # Expand to the left
                if s1_on_bd: print("** Returning on bd2"); return s1 #can't expand any further to left
                s2,g2 = s1,g1
                s1 = s3 - (s3-s1)*PHI; g1 = g(s1)
                if g1 is None:
                    s1,g1 = _findBoundary(g,s2,s1)
                    s1_on_bd = True

            else:  # Got it bracketed: now just narrow down bracket
                return _max_within_bracket(g,s1,g1,s2,g2,s3,g3)

    print("Warning: maximize_1d could not find bracket")

    ret = s2 if g2 is not None else s1 #return s2 if it evaluates to a valid point
    assert( g(ret) is not None )       #  otherwise return s1, since it should always be valid
    return ret


#TODO: g3 and g1 are unused!
def _max_within_bracket(g,s1,g1,s2,g2,s3,g3):
    TOL = 1e-10; FRAC_TOL = 1e-6
    assert( s2-s1 > TOL )  # a legit bracket must have s1 < s2
    while( (abs(s3-s1) > TOL) and (abs(s3-s1) > FRAC_TOL*(abs(s3)+abs(s1))) ):
        s4=s1+(s3-s2); g4 = g(s4)
        #print "DEBUG: Max in brk iter: (",s1,",",g1,") (",s2,",",g2,") (",s3,",",g3,") (",s4,",",g4,")"
        assert(g4 is not None) #assume function is defined at all points in bracket
        if s4 > s2:
            if( g4 > g2 ): # Drop x1 (move to x2), move x2 to x4.
                s1,g1 = s2,g2
                s2,g2 = s4,g4
            else: # Drop x3 (move to x4)
                s3,g3 = s4,g4
        else:
            if g4 > g2: # Drop x3 (move to x2), move x2 to x4
                s3,g3 = s2,g2
                s2,g2 = s4,g4
            else: # Drop x1 (move to x4)
                s1,g1 = s4,g4
    return s2


#find boundary of g (i.e. at the edge of where it is defined)
# g(s1) must be defined (not None) and g(s2) must == None (function undefined)
def _findBoundary(g,s1,s2):
    #print "DEBUG: finding bd fn"
    TOL = 1e-6
    while( abs(s1-s2) > TOL ): #just do binary search
        m = (s1+s2)/2.0; gm = g(m)
        if gm is None: s2 = m
        else: s1 = m
    return s1,g(s1)




#provide finite difference derivatives with boundary for a given function f.  Boundaries are
# determined by the function f returning a None value when it is not defined.
def _finite_diff_dfdx_and_bdflag(f,x,DELTA):
    x = x.copy() #make sure x is unaltered
    N = len(x)
    dfdx = _np.zeros(N) #complex?
    bd   = _np.zeros(N)
    for k in range(N):
        x[k] += DELTA;   fPlus  = f(x)
        x[k] -= 2*DELTA; fMinus = f(x)
        x[k] += DELTA
        if fPlus  is None: bd[k] = +1.0
        elif fMinus is None: bd[k] = -1.0
        else: dfdx[k] = (fPlus - fMinus)/(2*DELTA)
        #assert(fPlus is not None or fMinus is not None) #make sure we don't evaluate f somewhere it's completely undefined

    return dfdx,bd

#def f6(param):
#    '''Schaffer's F6 function'''
#    para = param*10
#    para = param[0:2]
#    num = (sin(sqrt((para[0] * para[0]) + (para[1] * para[1])))) * \
#        (sin(sqrt((para[0] * para[0]) + (para[1] * para[1])))) - 0.5
#    denom = (1.0 + 0.001 * ((para[0] * para[0]) + (para[1] * para[1]))) * \
#            (1.0 + 0.001 * ((para[0] * para[0]) + (para[1] * para[1])))
#    f6 =  0.5 - (num/denom)
#    errorf6 = 1 - f6
#    return f6, errorf6;



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Functions for selecting a complete set of germs for a GST analysis."""

import warnings as _warnings

import numpy as _np
import numpy.linalg as _nla

from .. import objects as _objs
from .. import construction as _constr
from . import grasp as _grasp
from . import scoring as _scoring


def generate_germs(gs_target, randomize=True, randomizationStrength=1e-2,
                   numGSCopies=5, seed=None, maxGermLength=6,
                   force="singletons", algorithm='greedy',
                   algorithm_kwargs=None, verbosity=1):
    """Generate a germ set for doing GST with a given target gateset.
    This function provides a streamlined interface to a variety of germ
    selection algorithms. It's goal is to provide a method that typical users
    can run by simply providing a target gateset and leaving all other settings
    at their default values, while providing flexibility for users desiring
    more control to fine tune some of the general and algorithm-specific
    details.
    Currently, to break troublesome degeneracies and provide some confidence
    that the chosen germ set is amplificationally complete (AC) for all
    gatesets in a neighborhood of the target gateset (rather than only the
    target gateset), an ensemble of gatesets with random unitary perturbations
    to their gates must be provided or generated.
    Parameters
    ----------
    gs_target : GateSet or list of GateSet
        The gateset you are aiming to implement, or a list of gatesets that are
        copies of the gateset you are trying to implement (either with or
        without random unitary perturbations applied to the gatesets).
    randomize : bool, optional
        Whether or not to add random unitary perturbations to the gateset(s)
        provided.
    randomizationStrength : float, optional
        The size of the random unitary perturbations applied to gates in the
        gateset. See :meth:`~pygsti.objects.GateSet.randomize_with_unitary`
        for more details.
    numGSCopies : int, optional
        The number of copies of the original gateset that should be used.
    seed : int, optional
        Seed for generating random unitary perturbations to gatesets. Also
        passed along to stochastic germ-selection algorithms.
    maxGermsLength : int, optional
        The maximum length (in terms of gates) of any germ allowed in the germ
        set. Currently will construct a list of all non-equivalent germs of
        length up to `maxGermsLength` for the germ selection algorithms to play
        around with.
    force : str or list, optional
        A list of GateStrings which *must* be included in the final germ set.
        If set to the special string "singletons" then all length-1 strings will
        be included.  Seting to None is the same as an empty list.
    algorithm : {'greedy', 'grasp', 'slack'}, optional
        Specifies the algorithm to use to generate the germ set. Current
        options are:
        'greedy'
            Add germs one-at-a-time until the set is AC, picking the germ that
            improves the germ-set score by the largest amount at each step. See
            :func:`build_up_breadth` for more details.
        'grasp'
            Use GRASP to generate random greedy germ sets and then locally
            optimize them. See :func:`grasp_germ_set_optimization` for more
            details.
        'slack'
            From a initial set of germs, add or remove a germ at each step in
            an attempt to improve the germ-set score. Will allow moves that
            degrade the score in an attempt to escape local optima as long as
            the degredation is within some specified amount of "slack". See
            :func:`optimize_integer_germs_slack` for more details.
    algorithm_kwargs : dict
        Dictionary of ``{'keyword': keyword_arg}`` pairs providing keyword
        arguments for the specified `algorithm` function. See the documentation
        for functions referred to in the `algorithm` keyword documentation for
        what options are available for each algorithm.
    verbosity : int, optional
        The verbosity level of the :class:`~pygsti.objects.VerbosityPrinter`
        used to print log messages.
    Returns
    -------
    list of GateString
        A list containing the germs making up the germ set.
    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)
    gatesetList = setup_gateset_list(gs_target, randomize,
                                     randomizationStrength, numGSCopies, seed)
    gates = gs_target.gates.keys()
    availableGermsList = _constr.list_all_gatestrings_without_powers_and_cycles(
        gates, maxGermLength)

    if algorithm_kwargs is None:
        # Avoid danger of using empty dict for default value.
        algorithm_kwargs = {}

    if algorithm == 'greedy':
        printer.log('Using greedy algorithm.', 1)
        # Define defaults for parameters that currently have no default or
        # whose default we want to change.
        default_kwargs = {
            'germsList': availableGermsList,
            'randomize': False,
            'seed': seed,
            'verbosity': max(0, verbosity - 1),
            'force': force,
            'scoreFunc': 'all',
            }
        for key in default_kwargs:
            if key not in algorithm_kwargs:
                algorithm_kwargs[key] = default_kwargs[key]
        germList = build_up_breadth(gatesetList=gatesetList,
                                    **algorithm_kwargs)
        if germList is not None:
            germsetScore = calculate_germset_score(
                germList, neighborhood=gatesetList,
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Constructed germ set:', 1)
            printer.log(str([str(germ) for germ in germList]), 1)
            printer.log('Score: {}'.format(germsetScore.score), 1)
    elif algorithm == 'grasp':
        printer.log('Using GRASP algorithm.', 1)
        # Define defaults for parameters that currently have no default or
        # whose default we want to change.
        default_kwargs = {
            'alpha': 0.1,   # No real reason for setting this value of alpha.
            'germsList': availableGermsList,
            'randomize': False,
            'seed': seed,
            'verbosity': max(0, verbosity - 1),
            'force': force,
            'returnAll': False,
            'scoreFunc': 'all',
            }
        for key in default_kwargs:
            if key not in algorithm_kwargs:
                algorithm_kwargs[key] = default_kwargs[key]
        germList = grasp_germ_set_optimization(gatesetList=gatesetList,
                                               **algorithm_kwargs)
        printer.log('Constructed germ set:', 1)

        if algorithm_kwargs['returnAll'] and germList[0] is not None:
            germsetScore = calculate_germset_score(
                germList[0], neighborhood=gatesetList,
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log(str([str(germ) for germ in germList[0]]), 1)
            printer.log('Score: {}'.format(germsetScore.score))
        elif not algorithm_kwargs['returnAll'] and germList is not None:
            germsetScore = calculate_germset_score(germList,
                                                   neighborhood=gatesetList)
            printer.log(str([str(germ) for germ in germList]), 1)
            printer.log('Score: {}'.format(germsetScore.score), 1)
    elif algorithm == 'slack':
        printer.log('Using slack algorithm.', 1)
        # Define defaults for parameters that currently have no default or
        # whose default we want to change.
        default_kwargs = {
            'germsList': availableGermsList,
            'randomize': False,
            'seed': seed,
            'verbosity': max(0, verbosity - 1),
            'force': force,
            'scoreFunc': 'all',
            }
        if ('slackFrac' not in algorithm_kwargs
                and 'fixedSlack' not in algorithm_kwargs):
            algorithm_kwargs['slackFrac'] = 0.1
        for key in default_kwargs:
            if key not in algorithm_kwargs:
                algorithm_kwargs[key] = default_kwargs[key]
        germList = optimize_integer_germs_slack(gatesetList,
                                                **algorithm_kwargs)
        if germList is not None:
            germsetScore = calculate_germset_score(
                germList, neighborhood=gatesetList,
                scoreFunc=algorithm_kwargs['scoreFunc'])
            printer.log('Constructed germ set:', 1)
            printer.log(str([str(germ) for germ in germList]), 1)
            printer.log('Score: {}'.format(germsetScore.score), 1)
    else:
        raise ValueError("'{}' is not a valid algorithm "
                         "identifier.".format(algorithm))

    return germList


def calculate_germset_score(germs, gs_target=None, neighborhood=None,
                            neighborhoodSize=5,
                            randomizationStrength=1e-2, scoreFunc='all',
                            gatePenalty=0.0, l1Penalty=0.0):
    """Calculate the score of a germ set with respect to a gate set.
    """
    scoreFn = lambda x: _scoring.list_score(x, scoreFunc=scoreFunc)
    if neighborhood is None:
        neighborhood = [gs_target.randomize_with_unitary(randomizationStrength)
                        for n in range(neighborhoodSize)]
    scores = [compute_non_AC_score(scoreFn, gateset=gateset,
                                   partialGermsList=germs,
                                   gatePenalty=gatePenalty,
                                   l1Penalty=l1Penalty)
              for gateset in neighborhood]

    return max(scores)


def get_gateset_params(gatesetList):
    """Get the number of gates and gauge parameters of the gatesets in a list.
    Also verify all gatesets have the same number of gates and gauge parameters.
    Parameters
    ----------
    gatesetList : list of GateSet
        A list of gatesets for which you want an AC germ set.
    Returns
    -------
    reducedGatesetList : list of GateSet
        The original list of gatesets with SPAM removed
    numGaugeParams : int
        The number of non-SPAM gauge parameters for all gatesets.
    numNonGaugeParams : int
        The number of non-SPAM non-gauge parameters for all gatesets.
    numGates : int
        The number of gates for all gatesets.
    Raises
    ------
    ValueError
        If the number of gauge parameters or gates varies among the gatesets.
    """
    # We don't care about SPAM, since it can't be amplified.
    reducedGatesetList = [removeSPAMVectors(gateset)
                          for gateset in gatesetList]

    # All the gatesets should have the same number of parameters and gates, but
    # let's be paranoid here for the time being and make sure.
    numGaugeParamsList = [reducedGateset.num_gauge_params()
                          for reducedGateset in reducedGatesetList]
    numGaugeParams = numGaugeParamsList[0]
    if not all([numGaugeParams == otherNumGaugeParams
                for otherNumGaugeParams in numGaugeParamsList[1:]]):
        raise ValueError("All gatesets must have the same number of gauge "
                         "parameters!")

    numNonGaugeParamsList = [reducedGateset.num_nongauge_params()
                             for reducedGateset in reducedGatesetList]
    numNonGaugeParams = numNonGaugeParamsList[0]
    if not all([numNonGaugeParams == otherNumNonGaugeParams
                for otherNumNonGaugeParams in numNonGaugeParamsList[1:]]):
        raise ValueError("All gatesets must have the same number of non-gauge "
                         "parameters!")

    numGatesList = [len(reducedGateset.gates)
                    for reducedGateset in reducedGatesetList]
    numGates = numGatesList[0]
    if not all([numGates == otherNumGates
                for otherNumGates in numGatesList[1:]]):
        raise ValueError("All gatesets must have the same number of gates!")

    return reducedGatesetList, numGaugeParams, numNonGaugeParams, numGates


def setup_gateset_list(gatesetList, randomize, randomizationStrength,
                       numCopies, seed):
    if not isinstance(gatesetList, (list, tuple)):
        gatesetList = [gatesetList]
    if len(gatesetList) > 1 and numCopies is not None:
        _warnings.warn("Ignoring numCopies={} since multiple gatesets were "
                       "supplied.".format(numCopies))
        assert(False)

    if randomize:
        gatesetList = randomizeGatesetList(gatesetList, randomizationStrength,
                                           numCopies, seed)

    return gatesetList


def compute_non_AC_score(scoreFn, thresholdAC=1e6, initN=1,
                         partialDerivDaggerDeriv=None, gateset=None,
                         partialGermsList=None, eps=None, numGaugeParams=None,
                         gatePenalty=0.0, germLengths=None, l1Penalty=0.0):
    """Compute the score for a germ set when it is not AC against a gateset.
    Normally scores computed for germ sets against gatesets for which they are
    not AC will simply be astronomically large. This is fine if AC is all you
    care about, but not so useful if you want to compare partial germ sets
    against one another to see which is closer to being AC. This function
    will see if the germ set is AC for the parameters corresponding to the
    largest `N` eigenvalues for increasing `N` until it finds a value of `N`
    for which the germ set is not AC or all the non gauge parameters are
    accounted for and report the value of `N` as well as the score.
    This allows partial germ set scores to be compared against one-another
    sensibly, where a larger value of `N` always beats a smaller value of `N`,
    and ties in the value of `N` are broken by the score for that value of `N`.
    Parameters
    ----------
    scoreFn : callable
        A function that takes as input a list of sorted eigenvalues and returns
        a score for the partial germ set based on those eigenvalues, with lower
        scores indicating better germ sets. Usually some flavor of
        :func:`~pygsti.algorithms.scoring.list_score`.
    thresholdAC : float, optional
        Value which the score (before penalties are applied) must be lower than
        for the germ set to be considered AC.
    initN : int
        The number of largest eigenvalues to begin with checking.
    partialDerivDaggerDeriv : numpy.array, optional
        Array with three axes, where the first axis indexes individual germs
        within the partial germ set and the remaining axes index entries in the
        positive square of the Jacobian of each individual germ's parameters
        with respect to the gateset parameters.
        If this array is not supplied it will need to be computed from
        `germsList` and `gateset`, which will take longer, so it is recommended
        to precompute this array if this routine will be called multiple times.
    gateset : GateSet, optional
        The gateset against which the germ set is to be scored. Not needed if
        `partialDerivDaggerDeriv` is provided.
    partialGermsList : list of GateString, optional
        The list of germs in the partial germ set to be evaluated. Not needed
        if `partialDerivDaggerDeriv` (and `germLengths` when
        ``gatePenalty > 0``) are provided.
    eps : float, optional
        Used when calculating `partialDerivDaggerDeriv` to determine if two
        eigenvalues are equal (see :func:`bulk_twirled_deriv` for details). Not
        used if `partialDerivDaggerDeriv` is provided.
    numGaugeParams : int
        The number of gauge parameters of the gateset. Not needed if `gateset`
        is provided.
    gatePenalty : float, optional
        Coefficient for a penalty linear in the sum of the germ lengths.
    germLengths : numpy.array, optional
        The length of each germ. Not needed if `gatePenalty` is ``0.0`` or
        `partialGermsList` is provided.
    l1Penalty : float, optional
        Coefficient for a penalty linear in the number of germs.
    Returns
    -------
    CompositeScore
        The score for the germ set indicating how many parameters it amplifies
        and its numerical score restricted to those parameters.
    """
    if partialDerivDaggerDeriv is None:
        if gateset is None or partialGermsList is None:
            raise ValueError("Must provide either partialDerivDaggerDeriv or "
                             "(gateset, partialGermsList)!")
        else:
            pDDD_kwargs = {'gateset': gateset, 'germsList': partialGermsList}
            if eps is not None:
                pDDD_kwargs['eps'] = eps
            if germLengths is not None:
                pDDD_kwargs['germLengths'] = germLengths
            partialDerivDaggerDeriv = calc_twirled_DDD(**pDDD_kwargs)

    if numGaugeParams is None:
        if gateset is None:
            raise ValueError("Must provide either numGaugeParams or gateset!")
        else:
            numGaugeParams = removeSPAMVectors(gateset).num_gauge_params()

    # Calculate penalty scores
    numGerms = partialDerivDaggerDeriv.shape[0]
    l1Score = l1Penalty*numGerms
    gateScore = 0.0
    if gatePenalty != 0.0:
        if germLengths is None:
            if partialGermsList is None:
                raise ValueError("Must provide either germLengths or "
                                 "partialGermsList when gatePenalty != 0.0!")
            else:
                germLengths = _np.array([len(germ)
                                         for germ in partialGermsList])
        gateScore = gatePenalty*_np.sum(germLengths)

    combinedDDD = _np.sum(partialDerivDaggerDeriv, axis=0)
    sortedEigenvals = _np.sort(_np.real(_nla.eigvalsh(combinedDDD)))
    observableEigenvals = sortedEigenvals[numGaugeParams:]
    N_AC = 0
    AC_score = _np.inf
    for N in range(initN, len(observableEigenvals) + 1):
        scoredEigenvals = observableEigenvals[-N:]
        candidate_AC_score = scoreFn(scoredEigenvals)
        if candidate_AC_score > thresholdAC:
            break   # We've found a set of parameters for which the germ set
                    # is not AC.
        else:
            AC_score = candidate_AC_score
            N_AC = N
    # Apply penalties
    score = AC_score + l1Score + gateScore

    return _scoring.CompositeScore(score, N_AC)


def calc_twirled_DDD(gateset, germsList, eps=None, check=False,
                     germLengths=None):
    """Calculate the positive squares of the germ Jacobians.
    twirledDerivDaggerDeriv == array J.H*J contributions from each germ
    (J=Jacobian) indexed by (iGerm, iGatesetParam1, iGatesetParam2)
    size (nGerms, vec_gateset_dim, vec_gateset_dim)
    """
    if germLengths is None:
        germLengths = _np.array([len(germ) for germ in germsList])
    btd_kwargs = {'gateset': gateset, 'gatestrings': germsList, 'check': check}
    if eps is not None:
        btd_kwargs['eps'] = eps
    twirledDeriv = bulk_twirled_deriv(**btd_kwargs)/germLengths[:, None, None]
    twirledDerivDaggerDeriv = _np.einsum('ijk,ijl->ikl',
                                         _np.conjugate(twirledDeriv),
                                         twirledDeriv)
    return twirledDerivDaggerDeriv


def compute_score(weights, gateset_num, scoreFunc, derivDaggerDerivList,
                  forceIndices, forceScore,
                  nGaugeParams, gatePenalty, germLengths, l1Penalty=1e-2,
                  scoreDict=None):
    """Returns a germ set "score" in which smaller is better.  Also returns
    intentionally bad score (`forceScore`) if `weights` is zero on any of
    the "forced" germs (i.e. at any index in `forcedIndices`).
    This function is included for use by :func:`optimize_integer_germs_slack`,
    but is not convenient for just computing the score of a germ set. For that,
    use :func:`calculate_germset_score`.
    """
    if forceIndices and _np.any(weights[forceIndices] <= 0):
        score = forceScore
    else:
        combinedDDD = _np.einsum('i,ijk', weights,
                                 derivDaggerDerivList[gateset_num])
        sortedEigenvals = _np.sort(_np.real(_nla.eigvalsh(combinedDDD)))
        observableEigenvals = sortedEigenvals[nGaugeParams:]
        score = (_scoring.list_score(observableEigenvals, scoreFunc)
                 + l1Penalty*_np.sum(weights)
                 + gatePenalty*_np.dot(germLengths, weights))
    if scoreDict is not None:
        # Side effect: calling compute_score caches result in scoreDict
        scoreDict[gateset_num, tuple(weights)] = score
    return score


def randomizeGatesetList(gatesetList, randomizationStrength, numCopies,
                         seed=None):
    if len(gatesetList) > 1 and numCopies is not None:
        raise ValueError("Input multiple gate sets XOR request multiple "
                         "copies only!")
    newgatesetList = []
    if len(gatesetList) > 1:
        for gatesetnum, gateset in enumerate(gatesetList):
            newgatesetList.append(gateset.randomize_with_unitary(
                randomizationStrength,
                seed=None if seed is None else seed+gatesetnum))
    else:
        for gatesetnum in range(numCopies if numCopies is not None else 1):
            newgatesetList.append(gatesetList[0].randomize_with_unitary(
                randomizationStrength,
                seed=None if seed is None else seed+gatesetnum))
    return newgatesetList


def checkGermsListCompleteness(gatesetList, germsList, scoreFunc, threshold):
    """Check to see if the germsList is amplificationally complete (AC)
    Checks for AC with respect to all the GateSets in `gatesetList`, returning
    the index of the first GateSet for which it is not AC or `-1` if it is AC
    for all GateSets.
    """
    for gatesetNum, gateset in enumerate(gatesetList):
        initial_test = test_germ_list_infl(gateset, germsList,
                                           scoreFunc=scoreFunc,
                                           threshold=threshold)
        if not initial_test:
            return gatesetNum

    # If the germsList is complete for all gatesets, return -1
    return -1


def removeSPAMVectors(gateset):
    reducedGateset = gateset.copy()
    for prepLabel in list(reducedGateset.preps.keys()):
        del reducedGateset.preps[prepLabel]
    for effectLabel in list(reducedGateset.effects.keys()):
        del reducedGateset.effects[effectLabel]
    return reducedGateset


def get_neighbors(boolVec):
    for i in range(len(boolVec)):
        v = boolVec.copy()
        v[i] = (v[i] + 1) % 2 # Toggle v[i] btwn 0 and 1
        yield v


def num_non_spam_gauge_params(gateset):
    """Return number of non-gauge, non-SPAM parameters in a GateSet.
    """
    return removeSPAMVectors(gateset).num_gauge_params()


# wrt is gate_dim x gate_dim, so is M, Minv, Proj
# so SOP is gate_dim^2 x gate_dim^2 and acts on vectorized *gates*
# Recall vectorizing identity (when vec(.) concats rows as flatten does):
#     vec( A * X * B ) = A tensor B^T * vec( X )
def _SuperOpForPerfectTwirl(wrt, eps):
    """Return super operator for doing a perfect twirl with respect to wrt.
    """
    assert wrt.shape[0] == wrt.shape[1] # only square matrices allowed
    dim = wrt.shape[0]
    SuperOp = _np.zeros((dim**2, dim**2), 'complex')

    # Get spectrum and eigenvectors of wrt
    wrtEvals, wrtEvecs = _np.linalg.eig(wrt)
    wrtEvecsInv = _np.linalg.inv(wrtEvecs)

    # We want to project  X -> M * (Proj_i * (Minv * X * M) * Proj_i) * Minv,
    # where M = wrtEvecs. So A = B = M * Proj_i * Minv and so
    # superop = A tensor B^T == A tensor A^T
    # NOTE: this == (A^T tensor A)^T while *Maple* germ functions seem to just
    # use A^T tensor A -> ^T difference
    for i in range(dim):
        # Create projector onto i-th eigenspace (spanned by i-th eigenvector
        # and other degenerate eigenvectors)
        Proj_i = _np.diag([(1 if (abs(wrtEvals[i] - wrtEvals[j]) <= eps)
                            else 0) for j in range(dim)])
        A = _np.dot(wrtEvecs, _np.dot(Proj_i, wrtEvecsInv))
        # Need to normalize, because we are overcounting projectors onto
        # subspaces of dimension d > 1, giving us d * Proj_i tensor Proj_i^T.
        # We can fix this with a division by tr(Proj_i) = d.
        SuperOp += _np.kron(A, A.T) / _np.trace(Proj_i)
        # SuperOp += _np.kron(A.T,A) # Mimic Maple version (but I think this is
        # wrong... or it doesn't matter?)
    return SuperOp  # a gate_dim^2 x gate_dim^2 matrix


def sq_sing_vals_from_deriv(deriv, weights=None):
    """Calculate the squared singulare values of the Jacobian of the germ set.
    Parameters
    ----------
    deriv : numpy.array
        Array of shape ``(nGerms, flattened_gate_dim, vac_gateset_dim)``. Each
        sub-array corresponding to an individual germ is the Jacobian of the
        vectorized gate representation of that germ raised to some power with
        respect to the gateset parameters, normalized by dividing by the length
        of each germ after repetition.
    weights : numpy.array
        Array of length ``nGerms``, giving the relative contributions of each
        individual germ's Jacobian to the combined Jacobian (which is calculated
        as a convex combination of the individual Jacobians).
    Returns
    -------
    numpy.array
        The sorted squared singular values of the combined Jacobian of the germ
        set.
    """
    # shape (nGerms, vec_gateset_dim, vec_gateset_dim)
    derivDaggerDeriv = _np.einsum('ijk,ijl->ikl', _np.conjugate(deriv), deriv)

    # Take the average of the D^dagger*D/L^2 matrices associated with each germ
    # with optional weights.
    combinedDDD = _np.average(derivDaggerDeriv, weights=weights, axis=0)
    sortedEigenvals = _np.sort(_np.real(_nla.eigvalsh(combinedDDD)))

    return sortedEigenvals


def twirled_deriv(gateset, gatestring, eps=1e-6):
    """Compute the "Twirled Derivative" of a gatestring.
    The twirled derivative is obtained by acting on the standard derivative of
    a gate string with the twirling superoperator.

    Parameters
    ----------
    gateset : Gateset object
        The GateSet which associates gate labels with operators.
    gatestring : GateString object
        The gate string to take a twirled derivative of.
    eps : float, optional
        Tolerance used for testing whether two eigenvectors are degenerate
        (i.e. abs(eval1 - eval2) < eps ? )

    Returns
    -------
    numpy array
      An array of shape (gate_dim^2, num_gateset_params)
    """
    prod = gateset.product(gatestring)

    # flattened_gate_dim x vec_gateset_dim
    dProd = gateset.dproduct(gatestring, flat=True)

    # flattened_gate_dim x flattened_gate_dim
    twirler = _SuperOpForPerfectTwirl(prod, eps)

    # flattened_gate_dim x vec_gateset_dim
    return _np.dot(twirler, dProd)


def bulk_twirled_deriv(gateset, gatestrings, eps=1e-6, check=False):
    """Compute the "Twirled Derivative" of a set of gatestrings.
    The twirled derivative is obtained by acting on the standard derivative of
    a gate string with the twirling superoperator.

    Parameters
    ----------
    gateset : Gateset object
        The GateSet which associates gate labels with operators.
    gatestrings : list of GateString objects
        The gate string to take a twirled derivative of.
    eps : float, optional
        Tolerance used for testing whether two eigenvectors are degenerate
        (i.e. abs(eval1 - eval2) < eps ? )
    check : bool, optional
        Whether to perform internal consistency checks, at the expense of
        making the function slower.

    Returns
    -------
    numpy array
        An array of shape (num_gate_strings, gate_dim^2, num_gateset_params)
    """
    evalTree = gateset.bulk_evaltree(gatestrings)
    dProds, prods = gateset.bulk_dproduct(evalTree, flat=True, bReturnProds=True)#, memLimit=None)
    gate_dim = gateset.get_dimension()
    fd = gate_dim**2 # flattened gate dimension

    ret = _np.empty((len(gatestrings), fd, dProds.shape[1]), 'complex')
    for i in range(len(gatestrings)):

        # flattened_gate_dim x flattened_gate_dim
        twirler = _SuperOpForPerfectTwirl(prods[i], eps)

        # flattened_gate_dim x vec_gateset_dim
        ret[i] = _np.dot(twirler, dProds[i*fd:(i+1)*fd])

    if check:
        for i, gatestring in enumerate(gatestrings):
            chk_ret = twirled_deriv(gateset, gatestring, eps)
            if _nla.norm(ret[i] - chk_ret) > 1e-6:
                _warnings.warn("bulk twirled derive norm mismatch = "
                               "%g - %g = %g"
                               % (_nla.norm(ret[i]), _nla.norm(chk_ret),
                                  _nla.norm(ret[i] - chk_ret)))

    return ret # nGateStrings x flattened_gate_dim x vec_gateset_dim



def test_germ_list_finitel(gateset, germsToTest, L, weights=None,
                           returnSpectrum=False, tol=1e-6):
    """Test whether a set of germs is able to amplify all non-gauge parameters.

    Parameters
    ----------
    gateset : GateSet
        The GateSet (associates gate matrices with gate labels).
    germsToTest : list of GateStrings
        List of germs gate sequences to test for completeness.
    L : int
        The finite length to use in amplification testing.  Larger
        values take longer to compute but give more robust results.
    weights : numpy array, optional
        A 1-D array of weights with length equal len(germsToTest),
        which multiply the contribution of each germ to the total
        jacobian matrix determining parameter amplification. If
        None, a uniform weighting of 1.0/len(germsToTest) is applied.
    returnSpectrum : bool, optional
        If True, return the jacobian^T*jacobian spectrum in addition
        to the success flag.
    tol : float, optional
        Tolerance: an eigenvalue of jacobian^T*jacobian is considered
        zero and thus a parameter un-amplified when it is less than tol.
    Returns
    -------
    success : bool
        Whether all non-gauge parameters were amplified.
    spectrum : numpy array
        Only returned when `returnSpectrum` is ``True``.  Sorted array of
        eigenvalues (from small to large) of the jacobian^T * jacobian
        matrix used to determine parameter amplification.
    """
    # Remove any SPAM vectors from gateset since we only want
    # to consider the set of *gate* parameters for amplification
    # and this makes sure our parameter counting is correct
    gateset = removeSPAMVectors(gateset)

    nGerms = len(germsToTest)
    germToPowL = [germ*L for germ in germsToTest]

    gate_dim = gateset.get_dimension()
    evt = gateset.bulk_evaltree(germToPowL)

    # shape (nGerms*flattened_gate_dim, vec_gateset_dim)
    dprods = gateset.bulk_dproduct(evt, flat=True)

    # shape (nGerms, flattened_gate_dim, vec_gateset_dim
    dprods = _np.reshape(dprods, (nGerms, gate_dim**2, dprods.shape[1]))

    germLengths = _np.array([len(germ) for germ in germsToTest], 'd')

    normalizedDeriv = dprods / (L * germLengths[:, None, None])

    sortedEigenvals = sq_sing_vals_from_deriv(normalizedDeriv, weights)

    nGaugeParams = gateset.num_gauge_params()

    observableEigenvals = sortedEigenvals[nGaugeParams:]

    bSuccess = bool(_scoring.list_score(observableEigenvals, 'worst') < 1/tol)

    return (bSuccess, sortedEigenvals) if returnSpectrum else bSuccess


def test_germ_list_infl(gateset, germsToTest, scoreFunc='all', weights=None,
                        returnSpectrum=False, threshold=1e6, check=False):
    """Test whether a set of germs is able to amplify all non-gauge parameters.

    Parameters
    ----------
    gateset : GateSet
        The GateSet (associates gate matrices with gate labels).
    germsToTest : list of GateString
        List of germs gate sequences to test for completeness.
    scoreFunc : string
        Label to indicate how a germ set is scored. See
        :func:`~pygsti.algorithms.scoring.list_score` for details.
    weights : numpy array, optional
        A 1-D array of weights with length equal len(germsToTest),
        which multiply the contribution of each germ to the total
        jacobian matrix determining parameter amplification. If
        None, a uniform weighting of 1.0/len(germsToTest) is applied.
    returnSpectrum : bool, optional
        If ``True``, return the jacobian^T*jacobian spectrum in addition
        to the success flag.
    threshold : float, optional
        An eigenvalue of jacobian^T*jacobian is considered zero and thus a
        parameter un-amplified when its reciprocal is greater than threshold.
        Also used for eigenvector degeneracy testing in twirling operation.
    check : bool, optional
      Whether to perform internal consistency checks, at the
      expense of making the function slower.
    Returns
    -------
    success : bool
        Whether all non-gauge parameters were amplified.
    spectrum : numpy array
        Only returned when `returnSpectrum` is ``True``.  Sorted array of
        eigenvalues (from small to large) of the jacobian^T * jacobian
        matrix used to determine parameter amplification.
    """
    # Remove any SPAM vectors from gateset since we only want
    # to consider the set of *gate* parameters for amplification
    # and this makes sure our parameter counting is correct
    gateset = removeSPAMVectors(gateset)


    germLengths = _np.array([len(germ) for germ in germsToTest], 'i')
    twirledDerivDaggerDeriv = calc_twirled_DDD(gateset, germsToTest,
                                               1./threshold, check,
                                               germLengths)
       # result[i] = _np.dot( twirledDeriv[i].H, twirledDeriv[i] ) i.e. matrix
       # product
       # result[i,k,l] = sum_j twirledDerivH[i,k,j] * twirledDeriv(i,j,l)
       # result[i,k,l] = sum_j twirledDeriv_conj[i,j,k] * twirledDeriv(i,j,l)

    if weights is None:
        nGerms = len(germsToTest)
        # weights = _np.array( [1.0/nGerms]*nGerms, 'd')
        weights = _np.array([1.0]*nGerms, 'd')

    combinedTDDD = _np.einsum('i,ijk', weights, twirledDerivDaggerDeriv)
    sortedEigenvals = _np.sort(_np.real(_np.linalg.eigvalsh(combinedTDDD)))

    nGaugeParams = gateset.num_gauge_params()
    observableEigenvals = sortedEigenvals[nGaugeParams:]

    bSuccess = bool(_scoring.list_score(observableEigenvals, scoreFunc)
                    < threshold)

    return (bSuccess, sortedEigenvals) if returnSpectrum else bSuccess


def build_up(gatesetList, germsList, randomize=True,
             randomizationStrength=1e-3, numCopies=None, seed=0, gatePenalty=0,
             scoreFunc='all', tol=1e-6, threshold=1e6, check=False,
             force="singletons", verbosity=0):
    """Greedy algorithm starting with 0 germs.
    Tries to minimize the number of germs needed to achieve amplificational
    completeness (AC). Begins with 0 germs and adds the germ that increases the
    score used to check for AC by the largest amount at each step, stopping when
    the threshold for AC is achieved.
    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    gatesetList = setup_gateset_list(gatesetList, randomize,
                                     randomizationStrength, numCopies, seed)

    (reducedGatesetList,
     numGaugeParams,
     numNonGaugeParams, numGates) = get_gateset_params(gatesetList)

    germLengths = _np.array([len(germ) for germ in germsList], 'i')
    numGerms = len(germsList)

    weights = _np.zeros(numGerms, 'i')
    goodGerms = []
    if force:
        if force == "singletons":
            weights[_np.where(germLengths == 1)] = 1
            goodGerms = [germ for germ
                             in _np.array(germsList)[_np.where(germLengths==1)]]
        else: #force should be a list of GateStrings
            for gs in force:
                weights[germsList.index(gs)] = 1
            goodGerms = force[:]

    undercompleteGatesetNum = checkGermsListCompleteness(gatesetList,
                                                         germsList,
                                                         scoreFunc,
                                                         threshold)
    if undercompleteGatesetNum > -1:
        printer.warning("Complete initial germ set FAILS on gateset "
                        + str(undercompleteGatesetNum) + ".")
        printer.warning("Aborting search.")
        return None

    printer.log("Complete initial germ set succeeds on all input gatesets.", 1)
    printer.log("Now searching for best germ set.", 1)

    printer.log("Starting germ set optimization. Lower score is better.", 1)

    twirledDerivDaggerDerivList = [calc_twirled_DDD(gateset, germsList, tol,
                                                    check, germLengths)
                                   for gateset in gatesetList]

    # Dict of keyword arguments passed to compute_score_non_AC that don't
    # change from call to call
    nonAC_kwargs = {
        'scoreFn': lambda x: _scoring.list_score(x, scoreFunc=scoreFunc),
        'thresholdAC': threshold,
        'numGaugeParams': numGaugeParams,
        'gatePenalty': gatePenalty,
        'germLengths': germLengths,
        }

    for gatesetNum, reducedGateset in enumerate(reducedGatesetList):
        derivDaggerDeriv = twirledDerivDaggerDerivList[gatesetNum]
        # Make sure the set of germs you come up with is AC for all
        # gatesets.
        # Remove any SPAM vectors from gateset since we only want
        # to consider the set of *gate* parameters for amplification
        # and this makes sure our parameter counting is correct
        while _np.any(weights == 0):
            # As long as there are some unused germs, see if you need to add
            # another one.
            if test_germ_list_infl(reducedGateset, goodGerms,
                                   scoreFunc=scoreFunc, threshold=threshold):
                # The germs are sufficient for the current gateset
                break
            candidateGerms = _np.where(weights == 0)[0]
            candidateGermScores = []
            for candidateGermIdx in _np.where(weights == 0)[0]:
                # If the germs aren't sufficient, try adding a single germ
                candidateWeights = weights.copy()
                candidateWeights[candidateGermIdx] = 1
                partialDDD = derivDaggerDeriv[
                    _np.where(candidateWeights == 1)[0], :, :]
                candidateGermScore = compute_non_AC_score(
                    partialDerivDaggerDeriv=partialDDD, **nonAC_kwargs)
                candidateGermScores.append(candidateGermScore)
            # Add the germ that give the best score
            bestCandidateGerm = candidateGerms[_np.array(
                candidateGermScores).argmin()]
            weights[bestCandidateGerm] = 1
            goodGerms.append(germsList[bestCandidateGerm])

    return goodGerms


def build_up_breadth(gatesetList, germsList, randomize=True,
                     randomizationStrength=1e-3, numCopies=None, seed=0,
                     gatePenalty=0, scoreFunc='all', tol=1e-6, threshold=1e6,
                     check=False, force="singletons", verbosity=0):
    """Greedy algorithm starting with 0 germs.
    Tries to minimize the number of germs needed to achieve amplificational
    completeness (AC). Begins with 0 germs and adds the germ that increases the
    score used to check for AC by the largest amount (for the gateset that
    currently has the lowest score) at each step, stopping when the threshold
    for AC is achieved. This strategy is something of a "breadth-first"
    approach, in contrast to :func:`build_up`, which only looks at the
    scores for one gateset at a time until that gateset achieves AC, then
    turning it's attention to the remaining gatesets.
    Parameters
    ----------
    germsList : list of GateString
        The list of germs to contruct a germ set from.
    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    gatesetList = setup_gateset_list(gatesetList, randomize,
                                     randomizationStrength, numCopies, seed)

    (reducedGatesetList,
     numGaugeParams,
     numNonGaugeParams, numGates) = get_gateset_params(gatesetList)

    germLengths = _np.array([len(germ) for germ in germsList], 'i')

    numGerms = len(germsList)

    goodGerms = []
    weights = _np.zeros(numGerms, 'i')
    if force:
        if force == "singletons":
            weights[_np.where(germLengths == 1)] = 1
            goodGerms = [germ for germ
                             in _np.array(germsList)[_np.where(germLengths==1)]]
        else: #force should be a list of GateStrings
            for gs in force:
                weights[germsList.index(gs)] = 1
            goodGerms = force[:]

    undercompleteGatesetNum = checkGermsListCompleteness(gatesetList,
                                                         germsList,
                                                         scoreFunc,
                                                         threshold)
    if undercompleteGatesetNum > -1:
        printer.warning("Complete initial germ set FAILS on gateset "
                        + str(undercompleteGatesetNum) + ".")
        printer.warning("Aborting search.")
        return None

    printer.log("Complete initial germ set succeeds on all input gatesets.", 1)
    printer.log("Now searching for best germ set.", 1)

    printer.log("Starting germ set optimization. Lower score is better.", 1)

    twirledDerivDaggerDerivList = [calc_twirled_DDD(gateset, germsList, tol,
                                                    check, germLengths)
                                   for gateset in gatesetList]

    # Dict of keyword arguments passed to compute_score_non_AC that don't
    # change from call to call
    nonAC_kwargs = {
        'scoreFn': lambda x: _scoring.list_score(x, scoreFunc=scoreFunc),
        'thresholdAC': threshold,
        'numGaugeParams': numGaugeParams,
        'gatePenalty': gatePenalty,
        'germLengths': germLengths,
        }


    initN = 1
    while _np.any(weights == 0):
        printer.log("Outer iteration: %d of %d amplified, %d germs" % 
                    (initN,numNonGaugeParams,len(goodGerms)), 2)
        # As long as there are some unused germs, see if you need to add
        # another one.
        if initN == numNonGaugeParams:
            break   # We are AC for all gatesets, so we can stop adding germs.

        candidateGerms = _np.where(weights == 0)[0]
        candidateGermScores = []
        candidateGermIndices = _np.where(weights == 0)[0]
        with printer.progress_logging(3):
            for i,candidateGermIdx in enumerate(candidateGermIndices):
                printer.show_progress(i, len(candidateGermIndices), 
                                      prefix="Inner iter over candidate germs",
                                      suffix=str(germsList[candidateGermIdx]))
    
                # If the germs aren't sufficient, try adding a single germ
                candidateWeights = weights.copy()
                candidateWeights[candidateGermIdx] = 1
                germVsGatesetScores = []
                for derivDaggerDeriv in twirledDerivDaggerDerivList:
                    # Loop over all gatesets
                    partialDDD = derivDaggerDeriv[
                        _np.where(candidateWeights == 1)[0], :, :]
                    germVsGatesetScores.append(compute_non_AC_score(
                        partialDerivDaggerDeriv=partialDDD, initN=initN,
                        **nonAC_kwargs))
                # Take the score for the current germ to be it's worst score over
                # all gatesets.
                worstScore = max(germVsGatesetScores)
                printer.log(str(worstScore), 4)
                candidateGermScores.append(worstScore)
        # Add the germ that gives the best worst score
        bestCandidateGerm = candidateGerms[_np.array(
            candidateGermScores).argmin()]
        weights[bestCandidateGerm] = 1
        goodGerms.append(germsList[bestCandidateGerm])
        bestScore = min(candidateGermScores)
        initN = bestScore.N
        printer.log("Added %s to final germs (%s)" % 
                    (str(germsList[bestCandidateGerm]), str(bestScore)), 3)

    return goodGerms


#@profile
def optimize_integer_germs_slack(gatesetList, germsList, randomize=True,
                                 randomizationStrength=1e-3, numCopies=None,
                                 seed=0, l1Penalty=1e-2, gatePenalty=0,
                                 initialWeights=None, scoreFunc='all',
                                 maxIter=100, fixedSlack=False,
                                 slackFrac=False, returnAll=False, tol=1e-6,
                                 check=False, force="singletons",
                                 forceScore=1e100, threshold=1e6,
                                 verbosity=1):
    """Find a locally optimal subset of the germs in germsList.
    Locally optimal here means that no single germ can be excluded
    without making the smallest non-gauge eigenvalue of the
    Jacobian.H*Jacobian matrix smaller, i.e. less amplified,
    by more than a fixed or variable amount of "slack", as
    specified by `fixedSlack` or `slackFrac`.
    Parameters
    ----------
    gatesetList : GateSet or list of GateSet
        The list of GateSets to be tested.  To ensure that the returned germ
        set is amplficationally complete, it is a good idea to score potential
        germ sets against a collection (~5-10) of similar gate sets.  The user
        may specify a single GateSet and a number of unitarily close copies to
        be made (set by the kwarg `numCopies`), or the user may specify their
        own list of GateSets, each of which in turn may or may not be
        randomized (set by the kwarg `randomize`).
    germsList : list of GateString
        List of all germs gate sequences to consider.
    randomize : Bool, optional
        Whether or not the input GateSet(s) are first subject to unitary
        randomization.  If ``False``, the user should perform the unitary
        randomization themselves.  Note:  If the GateSet(s) are perfect (e.g.
        ``std1Q_XYI.gs_target``), then the germ selection output should not be
        trusted, due to accidental degeneracies in the GateSet.  If the
        GateSet(s) include stochastic (non-unitary) error, then germ selection
        will fail, as we score amplificational completeness in the limit of
        infinite sequence length (so any stochastic noise will completely
        depolarize any sequence in that limit).  Default is ``True``.
    randomizationStrength : float, optional
        The strength of the unitary noise used to randomize input GateSet(s);
        is passed to :func:`~pygsti.objects.GateSet.randomize_with_unitary`.
        Default is ``1e-3``.
    numCopies : int, optional
        The number of GateSet copies to be made of the input GateSet (prior to
        unitary randomization).  If more than one GateSet is passed in,
        `numCopies` should be ``None``.  If only one GateSet is passed in and
        `numCopies` is ``None``, no extra copies are made.
    seed : float, optional
        The starting seed used for unitary randomization.  If multiple GateSets
        are to be randomized, ``gatesetList[i]`` is randomized with ``seed +
        i``.  Default is 0.
    l1Penalty : float, optional
        How strong the penalty should be for increasing the germ set list by a
        single germ.  Default is 1e-2.
    gatePenalty : float, optional
        How strong the penalty should be for increasing a germ in the germ set
        list by a single gate.  Default is 0.
    initialWeights : list-like
        List or array of either booleans or (0 or 1) integers
        specifying which germs in `germList` comprise the initial
        germ set.  If ``None``, then starting point includes all
        germs.
    scoreFunc : string
        Label to indicate how a germ set is scored. See
        :func:`~pygsti.algorithms.scoring.list_score` for details.
    maxIter : int, optional
        The maximum number of iterations before giving up.
    fixedSlack : float, optional
        If not ``None``, a floating point number which specifies that excluding
        a germ is allowed to increase 1.0/smallest-non-gauge-eigenvalue by
        `fixedSlack`.  You must specify *either* `fixedSlack` or `slackFrac`.
    slackFrac : float, optional
        If not ``None``, a floating point number which specifies that excluding
        a germ is allowed to increase 1.0/smallest-non-gauge-eigenvalue by
        `fixedFrac`*100 percent.  You must specify *either* `fixedSlack` or
        `slackFrac`.
    returnAll : bool, optional
        If ``True``, return the final ``weights`` vector and score dictionary
        in addition to the optimal germ list (see below).
    tol : float, optional
        Tolerance used for eigenvector degeneracy testing in twirling
        operation.
    check : bool, optional
        Whether to perform internal consistency checks, at the
        expense of making the function slower.
    force : str or list, optional
        A list of GateStrings which *must* be included in the final germ set.
        If set to the special string "singletons" then all length-1 strings will
        be included.  Seting to None is the same as an empty list.
    forceScore : float, optional (default is 1e100)
        When `force` designates a non-empty set of gate strings, the score to
        assign any germ set that does not contain each and every required germ.
    threshold : float, optional (default is 1e6)
        Specifies a maximum score for the score matrix, above which the germ
        set is rejected as amplificationally incomplete.
    verbosity : int, optional
        Integer >= 0 indicating the amount of detail to print.

    Returns
    -------
    finalGermList : list
        Sublist of `germList` specifying the final, optimal set of germs.
    weights : array
        Integer array, of length ``len(germList)``, containing 0s and 1s to
        indicate which elements of `germList` were chosen as `finalGermList`.
        Only returned when `returnAll` is ``True``.
    scoreDictionary : dict
        Dictionary with keys which are tuples of 0s and 1s of length
        ``len(germList)``, specifying a subset of germs, and values ==
        1.0/smallest-non-gauge-eigenvalue "scores".
    See Also
    --------
    :class:`~pygsti.objects.GateSet`
    :class:`~pygsti.objects.GateString`
    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    gatesetList = setup_gateset_list(gatesetList, randomize,
                                     randomizationStrength, numCopies, seed)

    if (fixedSlack and slackFrac) or (not fixedSlack and not slackFrac):
        raise ValueError("Either fixedSlack *or* slackFrac should be specified")

    if initialWeights is not None:
        if len(germsList) != len(initialWeights):
            raise ValueError("The lengths of germsList (%d) and "
                             "initialWeights (%d) must match."
                             % (len(germsList), len(initialWeights)))
        # Normalize the weights array to be 0s and 1s even if it is provided as
        # bools
        weights = _np.array([1 if x else 0 for x in initialWeights])
    else:
        weights = _np.ones(len(germsList), 'i') # default: start with all germs
#        lessWeightOnly = True # we're starting at the max-weight vector

    undercompleteGatesetNum = checkGermsListCompleteness(gatesetList,
                                                         germsList, scoreFunc,
                                                         threshold)
    if undercompleteGatesetNum > -1:
        printer.log("Complete initial germ set FAILS on gateset "
                    + str(undercompleteGatesetNum) + ".", 1)
        printer.log("Aborting search.", 1)
        return (None, None, None) if returnAll else None

    printer.log("Complete initial germ set succeeds on all input gatesets.", 1)
    printer.log("Now searching for best germ set.", 1)

    num_gatesets = len(gatesetList)

    # Remove any SPAM vectors from gateset since we only want
    # to consider the set of *gate* parameters for amplification
    # and this makes sure our parameter counting is correct
    gateset0 = removeSPAMVectors(gatesetList[0])

    # Initially allow adding to weight. -- maybe make this an argument??
    lessWeightOnly = False

    nGaugeParams = gateset0.num_gauge_params()

    # score dictionary:
    #   keys = (gatesetNum, tuple-ized weight vector of 1's and 0's only)
    #   values = list_score
    scoreD = {}
    germLengths = _np.array([len(germ) for germ in germsList], 'i')

    if force:
        if force == "singletons":
            forceIndices = _np.where(germLengths == 1)
        else: #force should be a list of GateStrings
            forceIndices = _np.array([germsList.index(gs) for gs in force])
    else:
        forceIndices = None

    twirledDerivDaggerDerivList = [calc_twirled_DDD(gateset, germsList, tol,
                                                    check, germLengths)
                                   for gateset in gatesetList]

    # Dict of keyword arguments passed to compute_score that don't change from
    # call to call
    cs_kwargs = {
        'scoreFunc': scoreFunc,
        'derivDaggerDerivList': twirledDerivDaggerDerivList,
        'forceIndices': forceIndices,
        'forceScore': forceScore,
        'nGaugeParams': nGaugeParams,
        'gatePenalty': gatePenalty,
        'germLengths': germLengths,
        'l1Penalty': l1Penalty,
        'scoreDict': scoreD,
        }

    scoreList = [compute_score(weights, gateset_num, **cs_kwargs)
                 for gateset_num in range(num_gatesets)]
    score = _np.max(scoreList)
    L1 = sum(weights) # ~ L1 norm of weights

    printer.log("Starting germ set optimization. Lower score is better.", 1)
    printer.log("Gateset has %d gauge params." % nGaugeParams, 1)

    with printer.progress_logging(1):
        for iIter in range(maxIter):
            printer.show_progress(iIter, maxIter,
                                  suffix="score=%g, nGerms=%d" % (score, L1))

            bFoundBetterNeighbor = False
            for neighbor in get_neighbors(weights):
                neighborScoreList = []
                for gateset_num in range(len(gatesetList)):
                    if (gateset_num, tuple(neighbor)) not in scoreD:
                        neighborL1 = sum(neighbor)
                        neighborScoreList.append(compute_score(neighbor,
                                                               gateset_num,
                                                               **cs_kwargs))
                    else:
                        neighborL1 = sum(neighbor)
                        neighborScoreList.append(scoreD[gateset_num,
                                                        tuple(neighbor)])

                neighborScore = _np.max(neighborScoreList)  # Take worst case.
                # Move if we've found better position; if we've relaxed, we
                # only move when L1 is improved.
                if neighborScore <= score and (neighborL1 < L1 or
                                               not lessWeightOnly):
                    weights, score, L1 = neighbor, neighborScore, neighborL1
                    bFoundBetterNeighbor = True

                    printer.log("Found better neighbor: "
                                "nGerms = %d score = %g" % (L1, score), 2)

            if not bFoundBetterNeighbor: # Time to relax our search.
                # From now on, don't allow increasing weight L1
                lessWeightOnly = True

                if fixedSlack is False:
                    # Note score is positive (for sum of 1/lambda)
                    slack = score*slackFrac
                    # print "slack =", slack
                else:
                    slack = fixedSlack
                assert slack > 0

                printer.log("No better neighbor. Relaxing score w/slack: "
                            + "%g => %g" % (score, score+slack), 2)
                # Artificially increase score and see if any neighbor is better
                # now...
                score += slack

                for neighbor in get_neighbors(weights):
                    scoreList = [scoreD[gateset_num, tuple(neighbor)]
                                 for gateset_num in range(len(gatesetList))]
                    maxScore = _np.max(scoreList)
                    if sum(neighbor) < L1 and maxScore < score:
                        weights, score, L1 = neighbor, maxScore, sum(neighbor)
                        bFoundBetterNeighbor = True
                        printer.log("Found better neighbor: "
                                    "nGerms = %d score = %g" % (L1, score), 2)

                if not bFoundBetterNeighbor: # Relaxing didn't help!
                    printer.log("Stationary point found!", 1)
                    break # end main for loop

            printer.log("Moving to better neighbor", 1)
            # print score
        else:
            printer.log("Hit max. iterations", 1)

    printer.log("score = %s" % score, 1)
    printer.log("weights = %s" % weights, 1)
    printer.log("L1(weights) = %s" % sum(weights), 1)

    goodGerms = []
    for index, val in enumerate(weights):
        if val == 1:
            goodGerms.append(germsList[index])

    if returnAll:
        return goodGerms, weights, scoreD
    else:
        return goodGerms


def germ_breadth_score_fn(germSet, germsList, twirledDerivDaggerDerivList,
                          nonAC_kwargs, initN=1):
    """Score a germ set against a collection of gatesets.
    Calculate the score of the germ set with respect to each member of a
    collection of gatesets and return the worst score among that collection.
    Parameters
    ----------
    germSet : list of GateString
        The set of germs to score.
    germsList : list of GateString
        The list of all germs whose Jacobians are provided in
        `twirledDerivDaggerDerivList`.
    twirledDerivDaggerDerivList : numpy.array
        Jacobians for all the germs in `germsList` stored as a 3-dimensional
        array, where the first index indexes the particular germ.
    nonAC_kwargs : dict
        Dictionary containing further arguments to pass to
        :func:`compute_non_AC_score` for the scoring of the germ set against
        individual gatesets.
    initN : int
        The number of eigenvalues to begin checking for amplificational
        completeness with respect to. Passed as an argument to
        :func:`compute_non_AC_score`.
    Returns
    -------
    CompositeScore
        The worst score over all gatesets of the germ set.
    """
    weights = _np.zeros(len(germsList))
    for germ in germSet:
        weights[germsList.index(germ)] = 1
    germsVsGatesetScores = []
    for derivDaggerDeriv in twirledDerivDaggerDerivList:
        # Loop over all gatesets
        partialDDD = derivDaggerDeriv[_np.where(weights == 1)[0], :, :]
        germsVsGatesetScores.append(compute_non_AC_score(
            partialDerivDaggerDeriv=partialDDD, initN=initN, **nonAC_kwargs))
    # Take the score for the current germ set to be its worst score over all
    # gatesets.
    return max(germsVsGatesetScores)


def grasp_germ_set_optimization(gatesetList, germsList, alpha, randomize=True,
                                randomizationStrength=1e-3, numCopies=None,
                                seed=None, l1Penalty=1e-2, gatePenalty=0.0,
                                scoreFunc='all', tol=1e-6, threshold=1e6,
                                check=False, force="singletons",
                                iterations=5, returnAll=False, shuffle=False,
                                verbosity=0):
    """Use GRASP to find a high-performing germ set.
    Parameters
    ----------
    gatesetList : GateSet or list of GateSet
        The list of GateSets to be tested.  To ensure that the returned germ
        set is amplficationally complete, it is a good idea to score potential
        germ sets against a collection (~5-10) of similar gate sets.  The user
        may specify a single GateSet and a number of unitarily close copies to
        be made (set by the kwarg `numCopies`, or the user may specify their
        own list of GateSets, each of which in turn may or may not be
        randomized (set by the kwarg `randomize`).
    germsList : list of GateString
        List of all germs gate sequences to consider.
    alpha : float
        A number between 0 and 1 that roughly specifies a score theshold
        relative to the spread of scores that a germ must score better than in
        order to be included in the RCL. A value of 0 for `alpha` corresponds
        to a purely greedy algorithm (only the best-scoring germ set is
        included in the RCL), while a value of 1 for `alpha` will include all
        germs in the RCL.
        See :func:`pygsti.algorithms.scoring.composite_rcl_fn` for more details.
    randomize : Bool, optional
        Whether or not the input GateSet(s) are first subject to unitary
        randomization.  If ``False``, the user should perform the unitary
        randomization themselves.  Note:  If the GateSet(s) are perfect (e.g.
        ``std1Q_XYI.gs_target``), then the germ selection output should not be
        trusted, due to accidental degeneracies in the GateSet.  If the
        GateSet(s) include stochastic (non-unitary) error, then germ selection
        will fail, as we score amplificational completeness in the limit of
        infinite sequence length (so any stochastic noise will completely
        depolarize any sequence in that limit).
    randomizationStrength : float, optional
        The strength of the unitary noise used to randomize input GateSet(s);
        is passed to :func:`~pygsti.objects.GateSet.randomize_with_unitary`.
        Default is ``1e-3``.
    numCopies : int, optional
        The number of GateSet copies to be made of the input GateSet (prior to
        unitary randomization).  If more than one GateSet is passed in,
        `numCopies` should be ``None``.  If only one GateSet is passed in and
        `numCopies` is ``None``, no extra copies are made.
    seed : float, optional
        The starting seed used for unitary randomization.  If multiple GateSets
        are to be randomized, ``gatesetList[i]`` is randomized with ``seed +
        i``.
    l1Penalty : float, optional
        How strong the penalty should be for increasing the germ set list by a
        single germ. Used for choosing between outputs of various GRASP
        iterations.
    gatePenalty : float, optional
        How strong the penalty should be for increasing a germ in the germ set
        list by a single gate.
    scoreFunc : string
        Label to indicate how a germ set is scored. See
        :func:`~pygsti.algorithms.scoring.list_score` for details.
    tol : float, optional
        Tolerance used for eigenvector degeneracy testing in twirling
        operation.
    threshold : float, optional (default is 1e6)
        Specifies a maximum score for the score matrix, above which the germ
        set is rejected as amplificationally incomplete.
    check : bool, optional
        Whether to perform internal consistency checks, at the
        expense of making the function slower.
    force : str or list, optional
        A list of GateStrings which *must* be included in the final germ set.
        If set to the special string "singletons" then all length-1 strings will
        be included.  Seting to None is the same as an empty list.
    iterations : int, optional
        The number of GRASP iterations to perform.
    returnAll : bool, optional
        Flag set to tell the routine if it should return lists of all
        initial constructions and local optimizations in addition to the
        optimal solution (useful for diagnostic purposes or if you're not sure
        what your `finalScoreFn` should really be).
    shuffle : bool, optional
        Whether the neighborhood should be presented to the optimizer in a
        random order (important since currently the local optimizer updates the
        solution to the first better solution it finds in the neighborhood).
    verbosity : int, optional
        Integer >= 0 indicating the amount of detail to print.
    Returns
    -------
    finalGermList : list of GateString
        Sublist of `germsList` specifying the final, optimal set of germs.
    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    gatesetList = setup_gateset_list(gatesetList, randomize,
                                     randomizationStrength, numCopies, seed)

    (reducedGatesetList,
     numGaugeParams,
     numNonGaugeParams, numGates) = get_gateset_params(gatesetList)

    germLengths = _np.array([len(germ) for germ in germsList], 'i')

    numGerms = len(germsList)

    initialWeights = _np.zeros(numGerms, dtype='i')
    if force:
        if force == "singletons":
            initialWeights[_np.where(germLengths == 1)] = 1
        else: #force should be a list of GateStrings
            for gs in force:
                initialWeights[germsList.index(gs)] = 1

    getNeighborsFn = lambda weights: _grasp.get_swap_neighbors(
        weights, forcedWeights=initialWeights, shuffle=shuffle)

    undercompleteGatesetNum = checkGermsListCompleteness(gatesetList,
                                                         germsList,
                                                         scoreFunc,
                                                         threshold)
    if undercompleteGatesetNum > -1:
        printer.warning("Complete initial germ set FAILS on gateset "
                        + str(undercompleteGatesetNum) + ".")
        printer.warning("Aborting search.")
        return None

    printer.log("Complete initial germ set succeeds on all input gatesets.", 1)
    printer.log("Now searching for best germ set.", 1)

    printer.log("Starting germ set optimization. Lower score is better.", 1)

    twirledDerivDaggerDerivList = [calc_twirled_DDD(gateset, germsList, tol,
                                                    check, germLengths)
                                   for gateset in gatesetList]

    # Dict of keyword arguments passed to compute_score_non_AC that don't
    # change from call to call
    nonAC_kwargs = {
        'scoreFn': lambda x: _scoring.list_score(x, scoreFunc=scoreFunc),
        'thresholdAC': threshold,
        'numGaugeParams': numGaugeParams,
        'gatePenalty': gatePenalty,
        'germLengths': germLengths,
        }

    final_nonAC_kwargs = nonAC_kwargs.copy()
    final_nonAC_kwargs['l1Penalty'] = l1Penalty

    scoreFn = (lambda germSet:
               germ_breadth_score_fn(germSet, germsList,
                                     twirledDerivDaggerDerivList, nonAC_kwargs,
                                     initN=1))
    finalScoreFn = (lambda germSet:
                    germ_breadth_score_fn(germSet, germsList,
                                          twirledDerivDaggerDerivList,
                                          final_nonAC_kwargs, initN=1))

    feasibleThreshold = _scoring.CompositeScore(threshold, numNonGaugeParams)

    rclFn = lambda x: _scoring.composite_rcl_fn(x, alpha)

    initialSolns = []
    localSolns = []

    for iteration in range(iterations):
        # This loop is parallelizable (each iteration is independent of all
        # other iterations).
        printer.log('Starting iteration {} of {}.'.format(iteration + 1,
                                                          iterations), 1)
        success = False
        failCount = 0
        while not success and failCount < 10:
            try:
                iterSolns = _grasp.do_grasp_iteration(
                    elements=germsList, greedyScoreFn=scoreFn, rclFn=rclFn,
                    localScoreFn=scoreFn,
                    getNeighborsFn=getNeighborsFn,
                    feasibleThreshold=feasibleThreshold,
                    initialElements=initialWeights, seed=seed,
                    verbosity=verbosity)

                initialSolns.append(iterSolns[0])
                localSolns.append(iterSolns[1])

                success = True
                printer.log('Finished iteration {} of {}.'.format(
                    iteration + 1, iterations), 1)
            except Exception as e:
                failCount += 1
                if failCount == 10:
                    raise e
                else:
                    printer.warning(e)

    finalScores = _np.array([finalScoreFn(localSoln)
                             for localSoln in localSolns])
    bestSoln = localSolns[_np.argmin(finalScores)]

    return (bestSoln, initialSolns, localSolns) if returnAll else bestSoln
